# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import json

import pytest
from click.testing import CliRunner
from flexmock import flexmock

from packit.api import PackitAPI
from packit.cli.builds import koji_build
from packit.cli.packit_base import packit_base
from packit.config import (
    package_config,
)
from packit.distgit import DistGit
from packit.local_project import LocalProject

DEFAULT_CONFIG_YAML = """
    {
        "specfile_path": "python-teamcity-messages.spec",
        "files_to_sync": ["python-teamcity-messages.spec", ".packit.yaml"],
        "upstream_package_name": "teamcity-messages",
        "downstream_package_name": "python-teamcity-messages",
        "upstream_project_url": "https://github.com/majamassarini/teamcity-messages",
        "upstream_tag_template": "v{version}",
        "jobs": [
            {
                "job": "copr_build",
                "trigger": "commit",
                "targets": ["fedora-rawhide", "fedora-stable"]
            },
            {
                "job": "propose_downstream",
                "trigger": "release",
                "dist_git_branches": ["fedora-rawhide", "fedora-stable"]
            },
            {
                "job": "koji_build",
                "trigger": "commit",
                "allowed_pr_authors": ["packit"],
                "dist_git_branches": ["fedora-rawhide", "fedora-stable"]
            },
            {
                "job": "bodhi_update",
                "trigger": "commit",
                "dist_git_branches": ["fedora-rawhide", "fedora-stable"]
            }
        ]
    }
    """


MONOREPO_CONFIG_YAML = """
    {
        "upstream_project_url": "https://github.com/majamassarini/teamcity-messages",
        "packages": {
            "python-teamcity-messages": {
                "downstream_package_name": "python-teamcity-messages",
                "paths": ["."],
                "specfile_path": "python-teamcity-messages.spec",
                "files_to_sync": ["python-teamcity-messages.spec", ".packit.yaml"],
                "upstream_package_name": "teamcity-messages",
                "upstream_tag_template": "v{version}"
            },
            "python-teamcity-messages-fake": {
                "downstream_package_name": "python-teamcity-messages-fake",
                "paths": ["."],
                "specfile_path": "python-teamcity-messages.spec",
                "files_to_sync": ["python-teamcity-messages.spec", ".packit.yaml"],
                "upstream_package_name": "teamcity-messages",
                "upstream_tag_template": "v{version}"
            }
        },
        "jobs": [
            {
                "job": "copr_build",
                "trigger": "commit",
                "targets": ["fedora-rawhide", "fedora-stable"],
                "packages": ["python-teamcity-messages-fake"]
            },
            {
                "job": "propose_downstream",
                "trigger": "release",
                "dist_git_branches": ["fedora-rawhide", "fedora-stable"],
                "packages": ["python-teamcity-messages-fake"]
            },
            {
                "job": "koji_build",
                "trigger": "commit",
                "allowed_pr_authors": ["packit"],
                "dist_git_branches": ["fedora-rawhide", "fedora-stable"],
                "packages": ["python-teamcity-messages-fake"]
            },
            {
                "job": "bodhi_update",
                "trigger": "commit",
                "dist_git_branches": ["fedora-rawhide", "fedora-stable"],
                "packages": ["python-teamcity-messages-fake"]
            }
        ]
    }
    """


@pytest.mark.parametrize(
    "package_config_yaml,how_many_builds",
    [
        pytest.param(DEFAULT_CONFIG_YAML, 1, id="default package config"),
        pytest.param(MONOREPO_CONFIG_YAML, 2, id="monorepo package config"),
    ],
)
def test_koji_build(package_config_yaml, how_many_builds):
    package_config_dict = json.loads(package_config_yaml)

    flexmock(package_config).should_receive("find_packit_yaml").and_return(
        flexmock(name=".packit.yaml", parent="/some/dir/teamcity-messages"),
    )
    flexmock(package_config).should_receive("load_packit_yaml").and_return(
        package_config_dict,
    )
    flexmock(LocalProject).should_receive("git_repo").and_return(
        flexmock(
            remotes=[],
            active_branch=flexmock(name="an active branch"),
            head=flexmock(is_detached=False),
        ),
    )
    # otherwise _dg and _local_project objects will be created
    # by debugger threads and you are not able to debug them
    flexmock(PackitAPI).should_receive("__repr__").and_return("")
    flexmock(DistGit).should_receive("__repr__").and_return("")
    flexmock(koji_build).should_receive("get_branches").and_return("rawhide")
    flexmock(PackitAPI).should_receive("init_kerberos_ticket").and_return()
    if how_many_builds == 1:
        flexmock(PackitAPI).should_receive("build").and_return().once()
    elif how_many_builds == 2:
        flexmock(PackitAPI).should_receive("build").and_return().twice()

    runner = CliRunner()
    runner.invoke(packit_base, ["build", "in-koji"])



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import distro
import pytest
from requre.cassette import DataTypes
from requre.modules_decorate_all_methods import (
    # record_git_module,
    record_requests_module,
    record_tempfile_module,
)

from packit.local_project import LocalProject
from tests_recording.testbase import PackitTest

PR_ID = "227"


@record_tempfile_module()
@record_requests_module
# @record_git_module()  # https://github.com/packit/requre/issues/232
class TestLocalProject(PackitTest):
    """Test LocalProject and save interaction with GitHub"""

    def cassette_setup(self, cassette):
        """requre requires this method to be present"""
        cassette.data_miner.data_type = DataTypes.Dict

    @staticmethod
    def commit_title(lp: LocalProject):
        commit_msg = lp.git_repo.head.commit.message
        return commit_msg.split("\n", 1)[0]

    @pytest.mark.skipif(
        distro.name() == "CentOS Stream" and distro.version() == "8",
        reason="This test is know to fail on el8: https://github.com/packit/requre/issues/233",
    )
    def test_checkout_pr(self):
        """Test PR checkout with and without merging"""
        project = LocalProject(
            git_project=self.project,
            pr_id=PR_ID,
            git_url=self._project_url,
            working_dir=self.static_tmp,
        )
        assert project.ref == f"pr/{PR_ID}"
        # check that HEAD of the merge matches HEAD of main
        main = project.git_repo.heads["main"]
        # 'Merge pull request #231 from packit/pre-commit-ci-update-config
        assert self.commit_title(project) == main.commit.message.split("\n", 1)[0]
        assert "koji_build" in (project.working_dir / ".packit.yaml").read_text()

    @pytest.mark.skipif(
        distro.name() == "CentOS Stream" and distro.version() == "8",
        reason="This test is know to fail on el8: https://github.com/packit/requre/issues/233",
    )
    def test_checkout_pr_no_merge(self):
        """Test PR checkout with and without merging"""
        project = LocalProject(
            git_project=self.project,
            pr_id=PR_ID,
            git_url=self._project_url,
            working_dir=self.static_tmp,
            merge_pr=False,
        )
        assert project.ref == f"pr/{PR_ID}"
        assert (
            self.commit_title(project)
            == "Run Koji builds automatically for new dist-git commits"
        )
        assert "koji_build" in (project.working_dir / ".packit.yaml").read_text()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import shlex
from logging import getLogger
from typing import Callable, Optional

from packit.actions import ActionName
from packit.command_handler import CommandHandler
from packit.config import MultiplePackages

logger = getLogger(__name__)


class ActionsHandler:
    def __init__(
        self,
        package_config: MultiplePackages,
        command_handler: CommandHandler,
    ):
        self.package_config = package_config
        self.command_handler = command_handler

    def run_action(
        self,
        actions: ActionName,
        method: Optional[Callable] = None,
        env: Optional[dict] = None,
        *args,
        **kwargs,
    ):
        """
        Run the method in the self._with_action block.

        Usage:

        >   self._run_action(
        >        action_name="sync", method=dg.sync_files, upstream_project=up.local_project
        >   )
        >   # If user provided custom command for the `sync`, it will be used.
        >   # Otherwise, the method `dg.sync_files` will be used
        >   # with parameter `upstream_project=up.local_project`
        >
        >   self._run_action(action_name="pre-sync")
        >   # This will be used as an optional hook

        :param actions: ActionName enum (Name of the action that can be overwritten
                                                in the package_config.actions)
        :param method: method to run if the action was not defined by user
                    (if not specified, the action can be used for custom hooks)
        :param args: args for the method
        :param kwargs: kwargs for the method
        """
        if not method:
            logger.debug(f"Running {actions} hook.")
        if self.with_action(action=actions, env=env) and method:
            method(*args, **kwargs)

    def has_action(self, action: ActionName) -> bool:
        """
        Is the action defined in the config?
        """
        return action in self.package_config.actions

    def get_commands_for_actions(self, action: ActionName) -> list[list[str]]:
        """
        Parse the following types of the structure and return list of commands in the form of list.

        I)
        action_name: "one cmd"

        II)
        action_name:
          - "one cmd""

        III)
        action_name:
          - ["one", "cmd"]


        Returns [["one", "cmd"]] for all of them.

        :param action: str or list[str] or list[list[str]]
        :return: list[list[str]]
        """
        configured_action = self.package_config.actions[action]
        if isinstance(configured_action, str):
            configured_action = [configured_action]

        if not isinstance(configured_action, list):
            raise ValueError(
                f"Expecting 'str' or 'list' as a command, got '{type(configured_action)}'. "
                f"The value: {configured_action}",
            )

        parsed_commands = []
        for cmd in configured_action:
            if isinstance(cmd, str):
                parsed_commands.append(shlex.split(cmd))
            elif isinstance(cmd, list):
                parsed_commands.append(cmd)
            else:
                raise ValueError(
                    f"Expecting 'str' or 'list' as a command, got '{type(cmd)}'. "
                    f"The value: {cmd}",
                )
        return parsed_commands

    def with_action(self, action: ActionName, env: Optional[dict] = None) -> bool:
        """
        If the action is defined in the self.package_config.actions,
        we run it and return False (so we can skip the if block)

        If the action is not defined, return True.

        Usage:

        >   if self._with_action(action_name="patch"):
        >       # Run default implementation
        >
        >   # Custom command was run if defined in the config

        Context manager is currently not possible without ugly hacks:
        https://stackoverflow.com/questions/12594148/skipping-execution-of-with-block
        https://www.python.org/dev/peps/pep-0377/ (rejected)

        :param action: ActionName enum (Name of the action that can be overwritten
                                                in the package_config.actions)
        :param env: dict with env vars to set for the command
        :return: True, if the action is not overwritten, False when custom command was run
        """
        logger.debug(f"Running {action}.")
        if action in self.package_config.actions:
            commands_to_run = self.get_commands_for_actions(action)
            logger.info(f"Using user-defined script for {action}: {commands_to_run}")
            for cmd in commands_to_run:
                self.command_handler.run_command(command=cmd, env=env, print_live=True)
            return False
        logger.debug(f"Running default implementation for {action}.")
        return True

    def get_output_from_action(
        self,
        action: ActionName,
        env: Optional[dict] = None,
    ) -> Optional[list[str]]:
        """
        Run self.actions[action] command(s) and return their outputs.
        """
        if action not in self.package_config.actions:
            return None

        commands_to_run = self.get_commands_for_actions(action)

        logger.info(f"Using user-defined script for {action}: {commands_to_run}")
        return [
            self.command_handler.run_command(
                cmd,
                return_output=True,
                env=env,
                print_live=True,
            ).stdout
            for cmd in commands_to_run
        ]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import subprocess
from pathlib import Path

import pytest
from flexmock import flexmock
from ogr import GithubService, GitlabService

from packit.local_project import LocalProject
from packit.utils.repo import create_new_repo
from tests.spellbook import initiate_git_repo


@pytest.mark.parametrize(
    "merge, hops_file_content",
    ((True, "Cascade\n"), (False, None)),
)
def test_pr_id_and_ref(tmp_path: Path, merge, hops_file_content):
    """p-s passes both ref and pr_id, we want to check out PR"""
    remote = tmp_path / "remote"
    remote.mkdir()
    create_new_repo(remote, ["--bare"])
    upstream_git = tmp_path / "upstream_git"
    upstream_git.mkdir()
    initiate_git_repo(upstream_git, push=True, upstream_remote=str(remote))
    # mimic github PR
    pr_id = "123"
    ref = (
        subprocess.check_output(["git", "rev-parse", "HEAD^"], cwd=upstream_git)
        .strip()
        .decode()
    )
    local_tmp_branch = "asdqwe"
    subprocess.check_call(["git", "branch", local_tmp_branch, ref], cwd=upstream_git)
    subprocess.check_call(
        ["git", "push", "origin", f"{local_tmp_branch}:refs/pull/{pr_id}/head"],
        cwd=upstream_git,
    )
    subprocess.check_call(["git", "branch", "-D", local_tmp_branch], cwd=upstream_git)

    git_project = flexmock(repo="random_name", namespace="random_namespace")
    git_project.should_receive("get_pr").and_return(flexmock(target_branch="main"))

    LocalProject(
        working_dir=upstream_git,
        offline=True,
        pr_id=pr_id,
        ref=ref,
        git_service=GithubService(),
        git_project=git_project,
        merge_pr=merge,
    )

    assert (
        subprocess.check_output(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
            cwd=upstream_git,
        )
        .strip()
        .decode()
        == f"pr/{pr_id}"
    )
    if hops_file_content:
        # the changes are merged into main so the file should be there
        assert (upstream_git / "hops").read_text() == hops_file_content
    else:
        # we haven't merged into main, so there should be no "hops" file
        assert not (upstream_git / "hops").is_file()


def test_pr_id_and_ref_gitlab(tmp_path: Path):
    """p-s passes both ref and pr_id, we want to check out PR"""
    remote = tmp_path / "remote"
    remote.mkdir()
    create_new_repo(remote, ["--bare"])
    upstream_git = tmp_path / "upstream_git"
    upstream_git.mkdir()
    initiate_git_repo(upstream_git, push=True, upstream_remote=str(remote))
    # mimic gitlab MR
    pr_id = "123"
    ref = (
        subprocess.check_output(["git", "rev-parse", "HEAD^"], cwd=upstream_git)
        .strip()
        .decode()
    )
    local_tmp_branch = "asdqwe"
    subprocess.check_call(["git", "branch", local_tmp_branch, ref], cwd=upstream_git)
    subprocess.check_call(
        [
            "git",
            "push",
            "origin",
            f"{local_tmp_branch}:refs/merge-requests/{pr_id}/head",
        ],
        cwd=upstream_git,
    )
    subprocess.check_call(["git", "branch", "-D", local_tmp_branch], cwd=upstream_git)

    git_project = flexmock(repo="random_name", namespace="random_namespace")
    git_project.should_receive("get_pr").and_return(flexmock(target_branch="main"))

    LocalProject(
        working_dir=upstream_git,
        offline=True,
        pr_id=pr_id,
        ref=ref,
        git_service=GitlabService(token="12345"),
        git_project=git_project,
    )

    assert (
        subprocess.check_output(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
            cwd=upstream_git,
        )
        .strip()
        .decode()
        == f"pr/{pr_id}"
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import re

import pytest
from flexmock import flexmock
from specfile import Specfile

from packit.config import CommonPackageConfig, Config, PackageConfig
from packit.constants import DISTGIT_HOSTNAME_CANDIDATES, EXISTING_BODHI_UPDATE_REGEX
from packit.distgit import DistGit
from packit.local_project import LocalProjectBuilder


@pytest.mark.parametrize(
    "target_branch, source_branch, prs, exists",
    [
        (
            "f31",
            "f31-update",
            [
                flexmock(
                    title="Update",
                    target_branch="f31",
                    source_branch="f31-update",
                    description="Upstream tag: 0.4.0\nUpstream commit: 6957453b",
                    author="packit",
                ),
            ],
            True,
        ),
        (
            "f32",
            "f31-update",
            [
                flexmock(
                    title="Update",
                    target_branch="f31",
                    source_branch="f31-update",
                    description="Upstream tag: 0.4.0\nUpstream commit: 6957453b",
                    author="packit",
                ),
            ],
            False,
        ),
        (
            "f31",
            "f31-update",
            [
                flexmock(
                    title="Update",
                    target_branch="f31",
                    description="Upstream tag: 0.4.0\nUpstream commit: 6957453b",
                    author="packit-stg",
                    source_branch="f31-update",
                ),
            ],
            False,
        ),
        (
            "f31",
            "f32-update",
            [
                flexmock(
                    title="Update",
                    target_branch="f31",
                    source_branch="f31-update",
                    description="Upstream tag: 0.4.0\nUpstream commit: 6957453b",
                    author="packit",
                ),
            ],
            False,
        ),
    ],
)
def test_existing_pr(git_repo_mock, target_branch, source_branch, prs, exists):
    user_mock = flexmock().should_receive("get_username").and_return("packit").mock()
    local_project = LocalProjectBuilder().build(
        git_project=flexmock(service="something", get_pr_list=lambda: prs),
        git_repo=git_repo_mock,
        git_service=flexmock(user=user_mock),
    )
    distgit = DistGit(
        config=flexmock(Config()),
        package_config=flexmock(
            PackageConfig(packages={"package": CommonPackageConfig()}),
        ),
        local_project=local_project,
    )
    pr = distgit.existing_pr(target_branch, source_branch)
    if exists:
        assert pr is not None
    else:
        assert pr is None


# Test covers the regression from monorepo refactoring that affects sync-release
# on downstream, since it directly accessed the attribute on the dist-git config
# instead of accessing specific package, which can cause ambiguity…
def test_monorepo_regression():
    config = flexmock(fas_user="mf")

    # Construct the package config; DON'T MOCK TO ENSURE IT CAN BE REPRODUCED
    package_a = flexmock(allowed_gpg_keys=["0xDEADBEEF"])
    package_b = flexmock(allowed_gpg_keys=["0xDEADBEEF"])
    package_config = PackageConfig(
        {
            "a": package_a,
            "b": package_b,
        },
    )

    dg = DistGit(config, package_config)

    # Assume the config has been synced to dist-git, therefore is 1:1 to the
    # one passed to DistGit class
    dg._downstream_config = package_config

    assert dg.get_allowed_gpg_keys_from_downstream_config() == ["0xDEADBEEF"]


# Test covers regex used for silencing of Bodhi exceptions for existing updates
@pytest.mark.parametrize(
    "exception_message, matches",
    [
        (
            (
                '{"status": "error", "errors": ['
                '{"location": "body", "name": "builds", "description": '
                '"Cannot find any tags associated with build: packit-0.79.1-1.el9"},'
                '{"location": "body", "name": "builds", "description": "Cannot '
                'find release associated with build: packit-0.79.1-1.el9, tags: []"}]}'
            ),
            False,
        ),
        (
            (
                '{"status": "error", "errors": ['
                '{"location": "body", "name": "builds", '
                '"description": "Update for linux-system-roles-1.53.4-1.fc39 already exists"}]}'
            ),
            True,
        ),
    ],
)
def test_bodhi_regex(exception_message, matches):
    assert bool(re.match(EXISTING_BODHI_UPDATE_REGEX, exception_message)) == matches


@pytest.mark.parametrize(
    "changelog, bugs",
    [
        (
            "* Fri Sep 29 2023 Packit <hello@packit.dev> - 0.82.0-1"
            "- Resolves rhbz#2240355",
            ["2240355"],
        ),
        (
            "* Fri Sep 29 2023 Packit <hello@packit.dev> - 0.82.0-1"
            "- Resolves rhbz#2240355"
            "- Resolves rhbz#2340355",
            ["2240355", "2340355"],
        ),
        (
            "* Fri Sep 29 2023 Packit <hello@packit.dev> - 0.82.0-1"
            "- Update without associated bugs",
            [],
        ),
    ],
)
def test_get_bugzilla_ids_from_changelog(changelog, bugs):
    assert DistGit.get_bugzilla_ids_from_changelog(changelog) == bugs


# Regression test for constructing the hostname candidates from the possible dist-git instances
def test_hostname_candidates():
    assert {
        "src.stg.fedoraproject.org",
        "gitlab.com",
        "src.fedoraproject.org",
        "pkgs.fedoraproject.org",
        "pkgs.stg.fedoraproject.org",
    } == DISTGIT_HOSTNAME_CANDIDATES


@pytest.mark.parametrize(
    "spec_source_id, with_coffee, archive_names",
    [
        ("Source0", "0", ["source0.tar.gz", "source2.tar.gz"]),
        ("Source1", "0", ["source0.tar.gz", "source1.tar.gz", "source2.tar.gz"]),
        ("Source2", "0", ["source0.tar.gz", "source2.tar.gz"]),
        ("Source2", "1", ["source0.tar.gz", "source2-with-coffee.tar.gz"]),
    ],
)
def test_upstream_archive_names(spec_source_id, with_coffee, archive_names, tmp_path):
    specfile_content = (
        "Name: test\n"
        "Version: 1.2.3\n"
        "Release: 1\n"
        "Source0: https://example.com/source0.tar.gz\n"
        "Source1: source1.tar.gz\n"
        "%if 0%{?with_coffee}\n"
        "Source2: https://example.com/source2-with-coffee.tar.gz\n"
        "%else\n"
        "Source2: https://example.com/source2.tar.gz\n"
        "%endif\n"
        "License: MIT\n"
        "Summary: test\n"
        "%description\ntest\n"
    )
    spec_path = tmp_path / "test.spec"
    spec_path.write_text(specfile_content)
    specfile = Specfile(
        spec_path,
        sourcedir=tmp_path,
        autosave=True,
        macros=[("with_coffee", with_coffee)],
    )
    dg = DistGit(
        config=flexmock(Config()),
        package_config=flexmock(
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="test.spec",
                        spec_source_id=spec_source_id,
                    ),
                },
            ),
        ),
    )
    flexmock(dg).should_receive("specfile").and_return(specfile)
    assert dg.upstream_archive_names == archive_names



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import tempfile
from pathlib import Path

import git
import pytest
from flexmock import flexmock

from packit import local_project
from packit.local_project import (
    CALCULATE,
    NOT_TO_CALCULATE,
    LocalProject,
    LocalProjectBuilder,
    LocalProjectCalculationState,
)
from packit.utils import repo
from packit.utils.repo import RepositoryCache


def test_parse_repo_name_and_namespace_from_namespace():
    project = LocalProject(full_name="namespace/repository_name", refresh=False)
    changed = project._parse_repo_name_full_name_and_namespace()

    assert changed
    assert project.repo_name == "repository_name"
    assert project.namespace == "namespace"
    assert project.full_name == "namespace/repository_name"


def test_parse_full_name_from_repo_and_namespace():
    project = LocalProject(
        repo_name="repository_name",
        namespace="namespace",
        refresh=False,
    )
    changed = project._parse_repo_name_full_name_and_namespace()

    assert changed
    assert project.repo_name == "repository_name"
    assert project.namespace == "namespace"
    assert project.full_name == "namespace/repository_name"


def test_parse_git_repo_from_working_dir():
    path = Path("some/example/path")
    flexmock(local_project).should_receive("is_git_repo").with_args(path).and_return(
        True,
    )
    flexmock(
        git,
        Repo=flexmock(active_branch="branch", head=flexmock(is_detached=False)),
    )
    project = LocalProject(working_dir=path, refresh=False)
    changed = project._parse_git_repo_from_working_dir()

    assert changed
    assert project.git_repo
    assert project.git_repo.active_branch == "branch"
    assert not project.working_dir_temporary


def test_parse_git_project_from_repo_namespace_and_git_service():
    service_mock = (
        flexmock()
        .should_receive("get_project")
        .with_args(repo="repo", namespace="namespace")
        .replace_with(lambda repo, namespace: flexmock())
        .mock()
    )

    project = LocalProject(
        git_service=service_mock,
        repo_name="repo",
        namespace="namespace",
        refresh=False,
    )
    changed = project._parse_git_project_from_repo_namespace_and_git_service()

    assert changed
    assert project.git_service
    assert project.git_project


def test_parse_git_service_from_git_project():
    """Get git_url, namespace/repo and service from git_project"""
    project_mock = flexmock(service=flexmock())

    project = LocalProject(git_project=project_mock, refresh=False)
    changed = project._parse_git_service_from_git_project()

    assert changed
    assert project.git_service
    assert project.git_project


def test_parse_ref_from_git_repo():
    """Get ref from git_repo"""
    project = LocalProject(
        git_repo=flexmock(
            active_branch=flexmock(name="branch"),
            head=flexmock(is_detached=False),
        ),
        refresh=False,
    )
    changed = project._parse_ref_from_git_repo()

    assert changed
    assert project.git_repo
    assert project._ref == "branch"


def test_parse_ref_from_git_repo_detached():
    project = LocalProject(
        git_repo=flexmock(
            active_branch="branch",
            head=flexmock(is_detached=True, commit=flexmock(hexsha="sha")),
        ),
        refresh=False,
    )
    changed = project._parse_ref_from_git_repo()

    assert changed
    assert project.git_repo
    assert project._ref == "sha"


def test_parse_git_repo_from_git_url():
    flexmock(local_project).should_receive("get_repo").with_args(
        url="http://some.example/url/reponame",
        directory=None,
    ).and_return(flexmock())
    project = LocalProject(git_url="http://some.example/url/reponame", refresh=False)
    changed = project._parse_git_repo_from_git_url()

    assert changed
    assert project.git_url
    assert project.git_repo
    assert project.working_dir_temporary

    project.working_dir_temporary = False


def test_parse_git_url_from_git_project():
    project = LocalProject(
        git_project=flexmock()
        .should_receive("get_git_urls")
        .and_return({"git": "http://some.example/namespace/reponame"})
        .once()
        .mock(),
        refresh=False,
    )

    changed = project._parse_git_url_from_git_project()

    assert changed
    assert project.git_project
    assert project.git_url == "http://some.example/namespace/reponame"


def test_parse_namespace_from_git_project():
    project = LocalProject(git_project=flexmock(namespace="namespace"), refresh=False)

    changed = project._parse_namespace_from_git_project()

    assert changed
    assert project.git_project
    assert project.namespace
    assert project.namespace == "namespace"


def test_parse_git_url_from_git_repo():
    project = LocalProject(
        git_repo=flexmock(
            remotes=[flexmock(name="origin", url="git@github.com:org/name")],
        ),
        refresh=False,
    )

    changed = project._parse_git_url_from_git_repo()

    assert changed
    assert project.git_repo
    assert project.git_url == "git@github.com:org/name"


def test_parse_namespace_from_git_url():
    project = LocalProject(
        git_url="https://github.com/namespace/reponame",
        refresh=False,
    )
    changed = project._parse_namespace_from_git_url()

    assert changed
    assert project.namespace
    assert project.namespace == "namespace"
    assert project.repo_name == "reponame"
    assert project.git_url == "https://github.com/namespace/reponame"


# CHECKOUT BRANCH


def test_clone_project_checkout_branch():
    """Checkout existing branch"""
    project = LocalProject(
        git_repo=flexmock(
            working_dir=Path("something"),
            git=flexmock().should_receive("checkout").with_args("other").once().mock(),
            commit=lambda: "9e131ba261d8d07fd1c55d8aff6ade085f5cd354",
        ),
        ref="other",
        git_url="git@github.com:org/name",
    )
    assert project.git_repo
    assert project.working_dir == Path("something")
    assert project._ref == "other"


# USAGE FROM CODE


def test_working_dir_namespace_repo_name():
    url = "https://server.git/my_namespace/package_name"
    work_dir = Path("./local/directory/to/git")
    flexmock(local_project).should_receive("is_git_repo").with_args(
        work_dir,
    ).and_return(True)

    flexmock(git).should_receive("Repo").with_args(path=work_dir).and_return(
        flexmock(
            active_branch=flexmock(name="branch"),
            head=flexmock(is_detached=False),
        ),
    )

    project = LocalProject(
        working_dir=work_dir,
        namespace="my_namespace",
        repo_name="package_name",
        git_service=flexmock()
        .should_receive("get_project")
        .and_return(flexmock(get_git_urls=lambda: {"git": url}))
        .mock(),
    )

    assert project
    assert project.git_url == url
    assert project.namespace == "my_namespace"
    assert project.repo_name == "package_name"
    assert project.git_service
    assert project.git_project
    assert project.git_repo
    assert project.ref == "branch"


def test_from_path_repo_name_git_service():
    flexmock(git.Repo).should_receive("clone_from").and_return(
        flexmock(
            active_branch=flexmock(name="branch"),
            working_dir="some/temp/dir",
            head=flexmock(is_detached=False),
        ),
    )

    flexmock(tempfile).should_receive("mkdtemp").and_return("some/temp/dir")

    project = LocalProject(
        git_url="https://server.git/my_namespace/package_name",
        repo_name="package_name",
        git_service=flexmock()
        .should_receive("get_project")
        .and_return(flexmock())
        .mock(),
    )

    assert project
    assert project.working_dir == Path("some/temp/dir")
    assert project.git_url == "https://server.git/my_namespace/package_name"
    assert project.namespace == "my_namespace"
    assert project.repo_name == "package_name"
    assert project.git_service
    assert project.git_project
    assert project.git_repo
    assert project.ref == "branch"


def test_working_dir():
    work_dir = Path("./local/directory/to/git")
    flexmock(local_project).should_receive("is_git_repo").with_args(
        work_dir,
    ).and_return(True)

    flexmock(git).should_receive("Repo").with_args(path=work_dir).and_return(
        flexmock(
            active_branch=flexmock(name="branch"),
            head=flexmock(is_detached=False),
            remotes=[
                flexmock(
                    name="origin",
                    url="https://server.git/my_namespace/package_name",
                ),
            ],
        ),
    )

    project = LocalProject(working_dir=work_dir)

    assert project
    assert project.git_url == "https://server.git/my_namespace/package_name"
    assert project.namespace == "my_namespace"
    assert project.working_dir == work_dir
    assert project.git_repo
    assert project.ref == "branch"


# OFFLINE


def test_offline_git_project():
    """No get_project on offline"""

    project = LocalProject(
        repo_name="repo_name",
        namespace="namespace",
        git_service=flexmock().should_receive("get_project").times(0).mock(),
        offline=True,
    )
    assert project.repo_name == "repo_name"
    assert project.namespace == "namespace"
    assert project.git_service
    assert not project.git_project


def test_offline_git_service():
    """No git service on on offline"""

    project = LocalProject(
        repo_name="repo_name",
        namespace="namespace",
        git_project=flexmock(service="something")
        .should_receive("get_git_urls")
        .times(0)
        .mock(),
        offline=True,
    )
    assert project.repo_name == "repo_name"
    assert project.namespace == "namespace"
    assert project.git_project
    assert not project.git_service
    assert not project.git_url


def test_offline_no_clone():
    """No clone on offline"""
    flexmock(repo).should_receive("get_repo").times(0)
    flexmock(tempfile).should_receive("mkdtemp").times(0)
    flexmock(git.Repo).should_receive("clone_from").times(0)

    project = LocalProject(
        working_dir="some/example/path",
        git_url="http://some.example/url/reponame",
        offline=True,
    )
    assert project.working_dir == Path("some/example/path")
    assert project.git_url == "http://some.example/url/reponame"
    assert not project.git_repo


def test_offline_no_clone_no_temp_dir():
    """No clone on offline, no temp dir"""
    flexmock(repo).should_receive("get_repo").times(0)
    flexmock(tempfile).should_receive("mkdtemp").times(0)
    flexmock(git.Repo).should_receive("clone_from").times(0)
    project = LocalProject(git_url="http://some.example/url/reponame", offline=True)

    assert project.git_url == "http://some.example/url/reponame"
    assert not project.git_repo
    assert not project.working_dir


def test_clone_using_empty_cache():
    cache_path_mock = flexmock(
        is_dir=lambda: True,
        iterdir=list,
        joinpath=lambda x: "/reference/repo/package_name",
    )
    repo_cache = RepositoryCache(cache_path=cache_path_mock, add_new=False)

    flexmock(git.Repo).should_receive("clone_from").with_args(
        url="https://server.git/my_namespace/package_name",
        to_path="some/temp/dir",
        tags=True,
    ).and_return(
        flexmock(
            active_branch=flexmock(name="branch"),
            working_dir="some/temp/dir",
            head=flexmock(is_detached=False),
        ),
    )

    flexmock(tempfile).should_receive("mkdtemp").and_return("some/temp/dir")

    project = LocalProject(
        git_url="https://server.git/my_namespace/package_name",
        repo_name="package_name",
        git_service=flexmock()
        .should_receive("get_project")
        .and_return(flexmock())
        .mock(),
        cache=repo_cache,
    )

    assert project
    assert project.working_dir == Path("some/temp/dir")
    assert project.git_url == "https://server.git/my_namespace/package_name"
    assert project.namespace == "my_namespace"
    assert project.repo_name == "package_name"
    assert project.git_service
    assert project.git_project
    assert project.git_repo
    assert project.ref == "branch"


def test_clone_using_cache_present():
    reference_repo = flexmock(
        is_dir=lambda: True,
        __str__="/reference/repo/package_name",
        name="package_name",
    )
    cache_path_mock = flexmock(
        is_dir=lambda: True,
        iterdir=lambda: [reference_repo],
        joinpath=lambda x: "/reference/repo/package_name",
    )
    repo_cache = RepositoryCache(cache_path=cache_path_mock, add_new=False)

    flexmock(git.Repo).should_receive("clone_from").with_args(
        url="https://server.git/my_namespace/package_name",
        reference="/reference/repo/package_name",
        to_path="some/temp/dir",
        tags=True,
    ).and_return(
        flexmock(
            active_branch=flexmock(name="branch"),
            working_dir="some/temp/dir",
            head=flexmock(is_detached=False),
        ),
    )

    flexmock(tempfile).should_receive("mkdtemp").and_return("some/temp/dir")

    project = LocalProject(
        git_url="https://server.git/my_namespace/package_name",
        repo_name="package_name",
        git_service=flexmock()
        .should_receive("get_project")
        .and_return(flexmock())
        .mock(),
        cache=repo_cache,
    )

    assert project
    assert project.working_dir == Path("some/temp/dir")
    assert project.git_url == "https://server.git/my_namespace/package_name"
    assert project.namespace == "my_namespace"
    assert project.repo_name == "package_name"
    assert project.git_service
    assert project.git_project
    assert project.git_repo
    assert project.ref == "branch"


def test_clone_using_cache_not_present():
    reference_repo = flexmock(
        is_dir=lambda: True,
        __str__="/reference/repo/different_package",
        name="different_package",
    )
    cache_path_mock = flexmock(
        is_dir=lambda: True,
        iterdir=lambda: [reference_repo],
        joinpath=lambda x: "/reference/repo/package_name",
    )
    repo_cache = RepositoryCache(cache_path=cache_path_mock, add_new=False)

    flexmock(git.Repo).should_receive("clone_from").with_args(
        url="https://server.git/my_namespace/package_name",
        to_path="some/temp/dir",
        tags=True,
    ).and_return(
        flexmock(
            active_branch=flexmock(name="branch"),
            working_dir="some/temp/dir",
            head=flexmock(is_detached=False),
        ),
    )

    flexmock(tempfile).should_receive("mkdtemp").and_return("some/temp/dir")

    project = LocalProject(
        git_url="https://server.git/my_namespace/package_name",
        repo_name="package_name",
        git_service=flexmock()
        .should_receive("get_project")
        .and_return(flexmock())
        .mock(),
        cache=repo_cache,
    )

    assert project
    assert project.working_dir == Path("some/temp/dir")
    assert project.git_url == "https://server.git/my_namespace/package_name"
    assert project.namespace == "my_namespace"
    assert project.repo_name == "package_name"
    assert project.git_service
    assert project.git_project
    assert project.git_repo
    assert project.ref == "branch"


def test_clone_and_add_to_cache():
    cache_path_mock = flexmock(
        is_dir=lambda: True,
        iterdir=list,
        joinpath=lambda x: "/reference/repo/package_name",
    )
    repo_cache = RepositoryCache(cache_path=cache_path_mock, add_new=True)

    flexmock(git.Repo).should_receive("clone_from").with_args(
        url="https://server.git/my_namespace/package_name",
        to_path="/reference/repo/package_name",
        tags=True,
    ).and_return(
        flexmock(
            active_branch=flexmock(name="branch"),
            working_dir="some/temp/dir",
            head=flexmock(is_detached=False),
        ),
    )
    flexmock(git.Repo).should_receive("clone_from").with_args(
        url="https://server.git/my_namespace/package_name",
        reference="/reference/repo/package_name",
        to_path="some/temp/dir",
        tags=True,
    ).and_return(
        flexmock(
            active_branch=flexmock(name="branch"),
            working_dir="some/temp/dir",
            head=flexmock(is_detached=False),
        ),
    )

    flexmock(tempfile).should_receive("mkdtemp").and_return("some/temp/dir")

    project = LocalProject(
        git_url="https://server.git/my_namespace/package_name",
        repo_name="package_name",
        git_service=flexmock()
        .should_receive("get_project")
        .and_return(flexmock())
        .mock(),
        cache=repo_cache,
    )

    assert project
    assert project.working_dir == Path("some/temp/dir")
    assert project.git_url == "https://server.git/my_namespace/package_name"
    assert project.namespace == "my_namespace"
    assert project.repo_name == "package_name"
    assert project.git_service
    assert project.git_project
    assert project.git_repo
    assert project.ref == "branch"


def test_builder_parse_repo_name_and_namespace_from_namespace():
    builder = LocalProjectBuilder()
    state = LocalProjectCalculationState(full_name="namespace/repository_name")
    changed = builder._parse_repo_name_full_name_and_namespace(state)

    assert changed
    assert state.repo_name == "repository_name"
    assert state.namespace == "namespace"
    assert state.full_name == "namespace/repository_name"


def test_builder_parse_full_name_from_repo_and_namespace():
    builder = LocalProjectBuilder()
    state = LocalProjectCalculationState(
        repo_name="repository_name",
        namespace="namespace",
    )
    changed = builder._parse_repo_name_full_name_and_namespace(state)

    assert changed
    assert state.repo_name == "repository_name"
    assert state.namespace == "namespace"
    assert state.full_name == "namespace/repository_name"


def test_builder_parse_git_repo_from_working_dir():
    path = Path("some/example/path")
    flexmock(local_project).should_receive("is_git_repo").with_args(path).and_return(
        True,
    )
    flexmock(
        git,
        Repo=flexmock(active_branch="branch", head=flexmock(is_detached=False)),
    )
    state = LocalProjectCalculationState(working_dir=path)
    builder = LocalProjectBuilder()
    changed = builder._parse_git_repo_from_working_dir(state)

    assert changed
    assert state.git_repo
    assert state.git_repo.active_branch == "branch"
    assert not state.working_dir_temporary


def test_builder_parse_git_project_from_repo_namespace_and_git_service():
    service_mock = (
        flexmock()
        .should_receive("get_project")
        .with_args(repo="repo", namespace="namespace")
        .replace_with(lambda repo, namespace: flexmock())
        .mock()
    )

    state = LocalProjectCalculationState(
        git_service=service_mock,
        repo_name="repo",
        namespace="namespace",
    )
    builder = LocalProjectBuilder()
    changed = builder._parse_git_project_from_repo_namespace_and_git_service(state)

    assert changed
    assert state.git_service
    assert state.git_project


def test_builder_parse_git_service_from_git_project():
    """Get git_url, namespace/repo and service from git_project"""
    project_mock = flexmock(service=flexmock())

    state = LocalProjectCalculationState(git_project=project_mock)
    changed = LocalProjectBuilder()._parse_git_service_from_git_project(state)

    assert changed
    assert state.git_service
    assert state.git_project


def test_builder_parse_ref_from_git_repo():
    """Get ref from git_repo"""
    state = LocalProjectCalculationState(
        git_repo=flexmock(
            active_branch=flexmock(name="branch"),
            head=flexmock(is_detached=False),
        ),
    )
    changed = LocalProjectBuilder()._parse_ref_from_git_repo(state)

    assert changed
    assert state.git_repo
    assert state.ref == "branch"


def test_builder_parse_ref_from_git_repo_detached():
    state = LocalProjectCalculationState(
        git_repo=flexmock(
            active_branch="branch",
            head=flexmock(is_detached=True, commit=flexmock(hexsha="sha")),
        ),
    )
    changed = LocalProjectBuilder()._parse_ref_from_git_repo(state)

    assert changed
    assert state.git_repo
    assert state.ref == "sha"


def test_builder_parse_working_dir_from_git_repo():
    state = LocalProjectCalculationState(git_repo=flexmock(working_dir="foo"))
    changed = LocalProjectBuilder()._parse_working_dir_from_git_repo(state)
    assert changed
    assert state.git_repo
    assert str(state.working_dir) == "foo"


def test_builder_parse_git_repo_from_git_url():
    flexmock(local_project).should_receive("get_repo").with_args(
        url="http://some.example/url/reponame",
        directory=None,
    ).and_return(flexmock())
    state = LocalProjectCalculationState(git_url="http://some.example/url/reponame")
    changed = LocalProjectBuilder()._parse_git_repo_from_git_url(state)

    assert changed
    assert state.git_url
    assert state.git_repo
    assert state.working_dir_temporary


def test_builder_parse_git_url_from_git_project():
    state = LocalProjectCalculationState(
        git_project=flexmock()
        .should_receive("get_git_urls")
        .and_return({"git": "http://some.example/namespace/reponame"})
        .once()
        .mock(),
    )

    changed = LocalProjectBuilder()._parse_git_url_from_git_project(state)

    assert changed
    assert state.git_project
    assert state.git_url == "http://some.example/namespace/reponame"


def test_builder_parse_repo_name_from_git_project():
    state = LocalProjectCalculationState(git_project=flexmock(repo="foo"))
    changed = LocalProjectBuilder()._parse_repo_name_from_git_project(state)

    assert changed
    assert state.git_project
    assert str(state.repo_name) == "foo"


def test_builder_parse_namespace_from_git_project():
    project = LocalProject(git_project=flexmock(namespace="namespace"), refresh=False)

    changed = project._parse_namespace_from_git_project()

    assert changed
    assert project.git_project
    assert project.namespace
    assert project.namespace == "namespace"


def test_builder_parse_git_url_from_git_repo():
    state = LocalProjectCalculationState(
        git_repo=flexmock(
            remotes=[flexmock(name="origin", url="git@github.com:org/name")],
        ),
    )

    changed = LocalProjectBuilder()._parse_git_url_from_git_repo(state)

    assert changed
    assert state.git_repo
    assert state.git_url == "git@github.com:org/name"


def test_builder_parse_namespace_from_git_url():
    state = LocalProjectCalculationState(
        git_url="https://github.com/namespace/reponame",
    )
    changed = LocalProjectBuilder()._parse_namespace_from_git_url(state)

    assert changed
    assert state.namespace
    assert state.namespace == "namespace"
    assert state.repo_name == "reponame"
    assert state.git_url == "https://github.com/namespace/reponame"


def test_builder_offline_git_project():
    """No get_project on offline"""

    builder = LocalProjectBuilder(offline=True)
    project = builder.build(
        repo_name="repo_name",
        namespace="namespace",
        git_service=flexmock().should_receive("get_project").times(0).mock(),
        git_project=CALCULATE,
    )
    assert project.repo_name == "repo_name"
    assert project.namespace == "namespace"
    assert project.git_service
    assert not project.git_project


def test_builder_offline_git_service():
    """No git service on on offline"""

    builder = LocalProjectBuilder(offline=True)
    project = builder.build(
        repo_name="repo_name",
        namespace="namespace",
        git_project=flexmock(service="something")
        .should_receive("get_git_urls")
        .times(0)
        .mock(),
        git_service=CALCULATE,
    )
    assert project.repo_name == "repo_name"
    assert project.namespace == "namespace"
    assert project.git_project
    assert not project.git_service
    assert not project.git_url


def test_builder_offline_no_clone():
    """No clone on offline"""
    flexmock(repo).should_receive("get_repo").times(0)
    flexmock(tempfile).should_receive("mkdtemp").times(0)
    flexmock(git.Repo).should_receive("clone_from").times(0)

    builder = LocalProjectBuilder(offline=True)
    project = builder.build(
        working_dir="some/example/path",
        git_url="http://some.example/url/reponame",
    )
    assert project.working_dir == Path("some/example/path")
    assert project.git_url == "http://some.example/url/reponame"
    assert not project.git_repo


def test_builder_offline_no_clone_no_temp_dir():
    """No clone on offline, no temp dir"""
    flexmock(repo).should_receive("get_repo").times(0)
    flexmock(tempfile).should_receive("mkdtemp").times(0)
    flexmock(git.Repo).should_receive("clone_from").times(0)
    project = LocalProject(git_url="http://some.example/url/reponame", offline=True)

    assert project.git_url == "http://some.example/url/reponame"
    assert not project.git_repo
    assert not project.working_dir


def test_builder_clone_using_empty_cache():
    cache_path_mock = flexmock(
        is_dir=lambda: True,
        iterdir=list,
        joinpath=lambda x: "/reference/repo/package_name",
    )
    repo_cache = RepositoryCache(cache_path=cache_path_mock, add_new=False)

    flexmock(git.Repo).should_receive("clone_from").with_args(
        url="https://server.git/my_namespace/package_name",
        to_path="some/temp/dir",
        tags=True,
    ).and_return(
        flexmock(
            active_branch=flexmock(name="branch"),
            working_dir="some/temp/dir",
            head=flexmock(is_detached=False),
        ),
    )

    flexmock(tempfile).should_receive("mkdtemp").and_return("some/temp/dir")

    builder = LocalProjectBuilder(cache=repo_cache)
    project = builder.build(
        git_url="https://server.git/my_namespace/package_name",
        repo_name="package_name",
        git_service=flexmock()
        .should_receive("get_project")
        .and_return(flexmock())
        .mock(),
        working_dir=CALCULATE,
        namespace=CALCULATE,
        git_project=CALCULATE,
        git_repo=CALCULATE,
    )

    assert project
    assert project.working_dir == Path("some/temp/dir")
    assert project.git_url == "https://server.git/my_namespace/package_name"
    assert project.namespace == "my_namespace"
    assert project.repo_name == "package_name"
    assert project.git_service
    assert project.git_project
    assert project.git_repo
    assert project.ref == "branch"


def test_builder_clone_using_cache_present():
    reference_repo = flexmock(
        is_dir=lambda: True,
        __str__="/reference/repo/package_name",
        name="package_name",
    )
    cache_path_mock = flexmock(
        is_dir=lambda: True,
        iterdir=lambda: [reference_repo],
        joinpath=lambda x: "/reference/repo/package_name",
    )
    repo_cache = RepositoryCache(cache_path=cache_path_mock, add_new=False)

    flexmock(git.Repo).should_receive("clone_from").with_args(
        url="https://server.git/my_namespace/package_name",
        reference="/reference/repo/package_name",
        to_path="some/temp/dir",
        tags=True,
    ).and_return(
        flexmock(
            active_branch=flexmock(name="branch"),
            working_dir="some/temp/dir",
            head=flexmock(is_detached=False),
        ),
    )

    flexmock(tempfile).should_receive("mkdtemp").and_return("some/temp/dir")

    builder = LocalProjectBuilder(cache=repo_cache)
    project = builder.build(
        git_url="https://server.git/my_namespace/package_name",
        repo_name="package_name",
        git_service=flexmock()
        .should_receive("get_project")
        .and_return(flexmock())
        .mock(),
        working_dir=CALCULATE,
        namespace=CALCULATE,
        git_project=CALCULATE,
        git_repo=CALCULATE,
    )

    assert project
    assert project.working_dir == Path("some/temp/dir")
    assert project.git_url == "https://server.git/my_namespace/package_name"
    assert project.namespace == "my_namespace"
    assert project.repo_name == "package_name"
    assert project.git_service
    assert project.git_project
    assert project.git_repo
    assert project.ref == "branch"


def test_builder_clone_using_cache_not_present():
    reference_repo = flexmock(
        is_dir=lambda: True,
        __str__="/reference/repo/different_package",
        name="different_package",
    )
    cache_path_mock = flexmock(
        is_dir=lambda: True,
        iterdir=lambda: [reference_repo],
        joinpath=lambda x: "/reference/repo/package_name",
    )
    repo_cache = RepositoryCache(cache_path=cache_path_mock, add_new=False)

    flexmock(git.Repo).should_receive("clone_from").with_args(
        url="https://server.git/my_namespace/package_name",
        to_path="some/temp/dir",
        tags=True,
    ).and_return(
        flexmock(
            active_branch=flexmock(name="branch"),
            working_dir="some/temp/dir",
            head=flexmock(is_detached=False),
        ),
    )

    flexmock(tempfile).should_receive("mkdtemp").and_return("some/temp/dir")

    builder = LocalProjectBuilder(cache=repo_cache)
    project = builder.build(
        git_url="https://server.git/my_namespace/package_name",
        repo_name="package_name",
        git_service=flexmock()
        .should_receive("get_project")
        .and_return(flexmock())
        .mock(),
        working_dir=CALCULATE,
        namespace=CALCULATE,
        git_project=CALCULATE,
        git_repo=CALCULATE,
    )

    assert project
    assert project.working_dir == Path("some/temp/dir")
    assert project.git_url == "https://server.git/my_namespace/package_name"
    assert project.namespace == "my_namespace"
    assert project.repo_name == "package_name"
    assert project.git_service
    assert project.git_project
    assert project.git_repo
    assert project.ref == "branch"


def test_builder_clone_and_add_to_cache():
    cache_path_mock = flexmock(
        is_dir=lambda: True,
        iterdir=list,
        joinpath=lambda x: "/reference/repo/package_name",
    )
    repo_cache = RepositoryCache(cache_path=cache_path_mock, add_new=True)

    flexmock(git.Repo).should_receive("clone_from").with_args(
        url="https://server.git/my_namespace/package_name",
        to_path="/reference/repo/package_name",
        tags=True,
    ).and_return(
        flexmock(
            active_branch=flexmock(name="branch"),
            working_dir="some/temp/dir",
            head=flexmock(is_detached=False),
        ),
    )
    flexmock(git.Repo).should_receive("clone_from").with_args(
        url="https://server.git/my_namespace/package_name",
        reference="/reference/repo/package_name",
        to_path="some/temp/dir",
        tags=True,
    ).and_return(
        flexmock(
            active_branch=flexmock(name="branch"),
            working_dir="some/temp/dir",
            head=flexmock(is_detached=False),
        ),
    )

    flexmock(tempfile).should_receive("mkdtemp").and_return("some/temp/dir")

    builder = LocalProjectBuilder(cache=repo_cache)
    project = builder.build(
        git_url="https://server.git/my_namespace/package_name",
        repo_name="package_name",
        git_service=flexmock()
        .should_receive("get_project")
        .and_return(flexmock())
        .mock(),
        working_dir=CALCULATE,
        namespace=CALCULATE,
        git_project=CALCULATE,
        git_repo=CALCULATE,
    )

    assert project
    assert project.working_dir == Path("some/temp/dir")
    assert project.git_url == "https://server.git/my_namespace/package_name"
    assert project.namespace == "my_namespace"
    assert project.repo_name == "package_name"
    assert project.git_service
    assert project.git_project
    assert project.git_repo
    assert project.ref == "branch"


def mock_no_parsing(builder: LocalProjectBuilder):
    flexmock(builder).should_receive("_parse_git_repo_from_working_dir").never()
    flexmock(builder).should_receive("_parse_repo_name_full_name_and_namespace").never()
    flexmock(builder).should_receive("_parse_git_repo_from_git_url").never()
    flexmock(builder).should_receive("_parse_git_project_from_url").never()
    flexmock(builder).should_receive(
        "_parse_git_project_from_repo_namespace_and_git_service",
    ).never()
    flexmock(builder).should_receive("_parse_git_service_from_git_project").never()
    flexmock(builder).should_receive("_parse_ref_from_git_repo").never()
    flexmock(builder).should_receive("_parse_git_url_from_git_project").never()
    flexmock(builder).should_receive("_parse_git_url_from_git_repo").never()
    flexmock(builder).should_receive("_parse_repo_name_from_git_project").never()
    flexmock(builder).should_receive("_parse_namespace_from_git_project").never()
    flexmock(builder).should_receive("_parse_namespace_from_git_url").never()


def test_builder_not_calculate_git_repo():
    cache_path_mock = flexmock(
        is_dir=lambda: True,
        iterdir=list,
        joinpath=lambda x: "/reference/repo/package_name",
    )
    repo_cache = RepositoryCache(cache_path=cache_path_mock, add_new=False)

    repo = "package_name"
    namespace = "my_namespace"
    git_url = "http://some.example/my_namespace/package_name"

    builder = LocalProjectBuilder(cache=repo_cache)
    project_mock = (
        flexmock(service=flexmock(), repo=repo, namespace=namespace)
        .should_receive("get_git_urls")
        .and_return({"git": git_url})
        .mock()
    )
    project = builder.build(
        git_project=project_mock,
        working_dir=Path(""),
        git_url=CALCULATE,
        repo_name=CALCULATE,
        namespace=CALCULATE,
        git_repo=NOT_TO_CALCULATE,
    )

    assert project
    assert project.git_url == git_url
    assert project.namespace == namespace
    assert project.repo_name == repo
    assert project.git_service
    assert project.git_project
    assert not project.git_repo


def test_builder_calculate_nothing():
    builder = LocalProjectBuilder()
    mock_no_parsing(builder)
    builder.build(working_dir="foo")


@pytest.fixture
def valid_repo():
    with tempfile.TemporaryDirectory() as tmpdir:
        result = git.Repo.init(tmpdir)
        yield result, tmpdir


def test_builder_direct_dependency(valid_repo):
    repo, path = valid_repo
    builder = LocalProjectBuilder()
    project = builder.build(git_repo=repo, working_dir=CALCULATE)
    assert str(project.working_dir) == path
    assert project.git_repo


def test_builder_transitive_prerequisite(valid_repo):
    repo, path = valid_repo
    repo.create_remote("origin", "https://github.com/packit/packit.git")
    builder = LocalProjectBuilder()
    project = builder.build(working_dir=path, git_url=CALCULATE)
    assert str(project.working_dir) == path
    assert project.git_repo
    assert project.git_url == "https://github.com/packit/packit.git"



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from packaging.version import InvalidVersion, parse


def compare_versions(a: str, b: str) -> int:
    """
    Parses and compares two versions.

    Args:
        a: Version A.
        b: Version B.

    Returns:
        Positive number if A > B, negative number if A < B, zero if they are equal.
    """
    try:
        v_a = parse(a)
    except InvalidVersion:
        v_a = None
    try:
        v_b = parse(b)
    except InvalidVersion:
        v_b = None
    if v_a is None:
        return -1 if v_b is not None else (a > b) - (a < b)
    if v_b is None:
        return 1
    return (v_a > v_b) - (v_a < v_b)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from contextlib import contextmanager
from pathlib import Path

import pytest

from packit.exceptions import PackitException

# from flexmock import flexmock
from packit.sync import SyncFilesItem, check_subpath


@contextmanager
def return_result(result):
    yield result


@pytest.mark.parametrize(
    "subpath,path,trailing_slash,result",
    [
        (
            Path("./test/this"),
            Path("."),
            False,
            return_result(str(Path("./test/this").resolve())),
        ),
        (
            Path("test/this"),
            Path("."),
            False,
            return_result(str(Path("test/this").resolve())),
        ),
        (
            Path("test/this"),
            Path("."),
            True,
            return_result(str(Path("test/this").resolve()) + "/"),
        ),
        (Path("../test/this"), Path("."), False, pytest.raises(PackitException)),
        (Path("test/../../this"), Path("."), False, pytest.raises(PackitException)),
    ],
)
def test_check_subpath(subpath, path, trailing_slash, result):
    with result as r:
        assert check_subpath(subpath, path, trailing_slash) == r


@pytest.mark.parametrize(
    "item,drop,result",
    [
        (SyncFilesItem(["a", "b"], "dest"), {"src": "b"}, SyncFilesItem(["a"], "dest")),
        (SyncFilesItem(["a"], "dest"), {"src": Path("a")}, None),
        (
            SyncFilesItem(["a", "b"], "dest"),
            {"src": "c"},
            SyncFilesItem(["a", "b"], "dest"),
        ),
        (
            SyncFilesItem(["src/a", "src/b"], "dest"),
            {"src": "a", "criteria": lambda x, y: Path(x).name == y},
            SyncFilesItem(["src/b"], "dest"),
        ),
        (
            SyncFilesItem(
                ["src/a"],
                "dest",
                filters=["dummy filter"],
                mkpath=True,
                delete=True,
            ),
            {"src": "c"},
            SyncFilesItem(
                ["src/a"],
                "dest",
                filters=["dummy filter"],
                mkpath=True,
                delete=True,
            ),
        ),
    ],
)
def test_drop_src(item, drop, result):
    """Check dropping a item from the src-list

    The 'src' argument can be a string or Path.
    When used in 'criteria' as 'y', there is no type conversion,
    so if the caller defines a custem 'criteria', it falls on them,
    to make sure the types of 'src' and 'y' match.
    """
    assert result == item.drop_src(**drop)


@pytest.mark.parametrize(
    "item,args,result",
    [
        (
            SyncFilesItem(["a", "b"], "dest"),
            {},
            return_result(
                SyncFilesItem(
                    [Path(Path.cwd() / "a"), Path(Path.cwd() / "b")],
                    Path(Path.cwd() / "dest"),
                ),
            ),
        ),
        (
            SyncFilesItem(["a", "b"], "dest"),
            {"src_base": Path("src").resolve(), "dest_base": Path("dest")},
            return_result(
                SyncFilesItem(
                    [Path(Path.cwd() / "src" / "a"), Path(Path.cwd() / "src" / "b")],
                    Path(Path.cwd() / "dest" / "dest"),
                ),
            ),
        ),
        (
            SyncFilesItem(["../a", "b"], "dest"),
            {"src_base": Path("src"), "dest_base": Path("dest").resolve()},
            pytest.raises(PackitException),
        ),
    ],
)
def test_resolve(item, args, result):
    with result as r:
        item.resolve(**args)
        assert r == item


@pytest.mark.parametrize(
    "item,args,result",
    [
        (
            SyncFilesItem(["a", "b"], "dest"),
            {},
            ["rsync", "--archive", "--ignore-missing-args", "a", "b", "dest"],
        ),
        (
            SyncFilesItem(["a", "b"], "dest2"),
            {"fail_on_missing": True},
            ["rsync", "--archive", "a", "b", "dest2"],
        ),
        (
            SyncFilesItem(["c", "d"], "dest", mkpath=True, delete=True),
            {},
            [
                "rsync",
                "--archive",
                "--delete",
                "--mkpath",
                "--ignore-missing-args",
                "c",
                "d",
                "dest",
            ],
        ),
        (
            SyncFilesItem(["c/*"], "dest", mkpath=True),
            {},
            ["rsync", "--archive", "--mkpath", "--ignore-missing-args", "c/*", "dest"],
        ),
        (
            SyncFilesItem(
                ["src/"],
                "dest",
                delete=True,
                filters=["protect .git*", "protect sources"],
            ),
            {"fail_on_missing": True},
            [
                "rsync",
                "--archive",
                "--delete",
                "--filter",
                "protect .git*",
                "--filter",
                "protect sources",
                "src/",
                "dest",
            ],
        ),
    ],
)
def test_command(item, args, result):
    assert result == item.command(**args)


def test_command_globs(tmp_path):
    src_dir = tmp_path / "src"
    src_dir.mkdir()
    (src_dir / "file1").touch()
    (src_dir / "file2").touch()
    item = SyncFilesItem(["src/*"], "dest")
    item.resolve(src_base=tmp_path, dest_base=tmp_path)
    command = item.command()
    # The return value of glob.glob() is unordered
    command = command[:-3] + sorted(command[-3:-1]) + command[-1:]
    assert [
        "rsync",
        "--archive",
        "--ignore-missing-args",
        f"{tmp_path}/src/file1",
        f"{tmp_path}/src/file2",
        f"{tmp_path}/dest",
    ] == command


def test_sync_files_item_sorting():
    order1 = [
        SyncFilesItem(src=["packit.spec"], dest="packit.spec"),
        SyncFilesItem(src=[".packit.yaml"], dest=".packit2.yaml"),
    ]
    order2 = [
        SyncFilesItem(src=[".packit.yaml"], dest=".packit2.yaml"),
        SyncFilesItem(src=["packit.spec"], dest="packit.spec"),
    ]
    assert sorted(order1) == sorted(order2)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
from pathlib import Path
from typing import Any, Union

import requests
from marshmallow import ValidationError

from packit.config import JobType
from packit.config.package_config import PackageConfig, get_local_specfile_path
from packit.constants import (
    ANITYA_MONITORING_CHECK_URL,
    RELEASE_MONITORING_PACKAGE_CHECK_URL,
)
from packit.exceptions import PackitConfigException
from packit.sync import iter_srcs

logger = logging.getLogger(__name__)


class PackageConfigValidator:
    """Validate the content of package configuration file (for example .packit.yaml)

    Attributes:
        config_file_path: Path of the configuration file to be validated.
        content: The content of the configuration file.
        project_path: Path of the project to which the configuration file belongs.
    """

    def __init__(
        self,
        config_file_path: Path,
        config_content: dict,
        project_path: Path,
        offline: bool = False,
    ):
        self.config_file_path = config_file_path
        self.content = config_content
        self.project_path = project_path
        self.offline = offline

    def validate(self) -> str:
        """Validate PackageConfig.

        Returns:
            String that the config is valid.

        Raises:
            PackitConfigException: when the config is not valid
        """
        schema_errors: Union[list[Any], dict[Any, Any]] = None
        config = None
        try:
            config = PackageConfig.get_from_dict(
                self.content,
                config_file_path=str(self.config_file_path),
                search_specfile=get_local_specfile_path,
                dir=self.config_file_path.parent,
                repo_name=self.project_path.name,
            )
        except ValidationError as e:
            schema_errors = e.messages

        specfile_path = self.content.get("specfile_path", None)
        if specfile_path and not (self.project_path / specfile_path).is_file():
            logger.warning(
                f"The spec file you defined ({specfile_path}) is not "
                f"present in the repository. If it's being generated "
                f"dynamically, you can verify the functionality by "
                f"running `packit srpm` to create an SRPM "
                f"from the current checkout. If it's not being generated, "
                f"please make sure the path is correct and "
                f"the file is present.",
            )

        synced_files_errors = []
        if config:
            for package_config in config.get_package_config_views().values():
                synced_files_errors = [
                    f
                    for f in iter_srcs(package_config.files_to_sync)
                    if not (
                        (self.project_path / package_config.paths[0] / f).exists()
                        or any(self.project_path.glob(f))
                    )
                ]  # right now we use just the first path in a monorepo package

                if (
                    any(
                        job
                        for job in package_config.get_job_views()
                        if job.type == JobType.pull_from_upstream
                    )
                ) and not self.offline:
                    package_name = package_config.downstream_package_name
                    self.check_upstream_release_monitoring_mapping_exists(package_name)
                    self.check_anitya_monitoring_enabled(package_name)

        output = f"{self.config_file_path.name} does not pass validation:\n"

        if schema_errors:
            if isinstance(schema_errors, list):
                output += "\n".join(map(str, schema_errors))
            else:
                for field_name, errors in schema_errors.items():
                    output += self.validate_get_field_output(errors, field_name)

        if synced_files_errors:
            output += (
                "The following {} configured to be synced but "
                "{} not present in the repository: {}\n"
            ).format(
                *(
                    (
                        "paths are",
                        "are",
                    )
                    if (len(synced_files_errors) > 1)
                    else ("path is", "is")
                ),
                ", ".join(synced_files_errors),
            )

        if schema_errors or synced_files_errors:
            raise PackitConfigException(output)
        return f"{self.config_file_path.name} is valid and ready to be used"

    def validate_get_field_output(
        self,
        errors: Union[list, dict],
        field_name: str,
        field_category: str = "* field",
        level: int = 1,
    ) -> str:
        if isinstance(errors, list):
            return f"{field_category} {field_name}: {errors[0]}\n"
        return self.validate_get_field_item_output(errors, field_name, level)

    def validate_get_field_item_output(
        self,
        errors: dict,
        field_name: str,
        level: int,
    ) -> str:
        index_output = f"{level * '*'} field {field_name} has an incorrect value:\n"
        level += 1
        for index, item_errors in errors.items():
            index_type_error = self.validate_get_field_output(
                item_errors,
                index,
                field_category=f"{level * '*'} value at index",
                level=level,
            )
            index_output += index_type_error
        return index_output

    @staticmethod
    def check_upstream_release_monitoring_mapping_exists(package_name: str):
        """
        Check whether there is mapping for the particular package in Upstream
        Release Monitoring and warn in case not.
        """
        try:
            response = requests.get(
                RELEASE_MONITORING_PACKAGE_CHECK_URL.format(package_name=package_name),
            )
            result = response.json()
            items = result.get("items")

            if not items:
                logger.warning(
                    f"No mapping for package {package_name!r} found in Upstream "
                    f"Release Monitoring. Please visit https://release-monitoring.org/ "
                    f"and create one, otherwise `pull_from_upstream` job won't be triggered.",
                )

        except requests.exceptions.RequestException as e:
            logger.error(
                f"Error while checking Upstream Release Monitoring "
                f"mapping for package {package_name!r}: {e}",
            )

    @staticmethod
    def check_anitya_monitoring_enabled(package_name: str):
        """
        Check whether the monitoring for the particular package is enabled
        and warn in case not.
        """
        try:
            response = requests.get(
                ANITYA_MONITORING_CHECK_URL.format(package_name=package_name),
            )
            result = response.json()
            if result.get("monitoring") == "no-monitoring":
                logger.warning(
                    f"Monitoring for package {package_name!r} is disabled. Please, visit "
                    f"https://src.fedoraproject.org/rpms/{package_name} and "
                    f"set `Monitoring status` on the left side to `Monitoring`, "
                    f"otherwise `pull_from_upstream` job won't be triggered.",
                )
            if result.get("monitoring") == "monitoring-with-scratch":
                logger.warning(
                    f"Monitoring for package {package_name!r} is set to "
                    f"`Monitoring and scratch builds`. Please, visit "
                    f"https://src.fedoraproject.org/rpms/{package_name} and "
                    f"set it to `Monitoring` on the left side (Monitoring status) "
                    f"to avoid duplicated scratch builds for new releases.",
                )

        except requests.exceptions.RequestException as e:
            logger.error(
                f"Error while checking monitoring for package {package_name!r}: {e}",
            )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from packit.constants import DISTGIT_INSTANCES
from packit.dist_git_instance import DistGitInstance

FEDORA_DG = DistGitInstance(
    hostname="src.fedoraproject.org",
    alternative_hostname="pkgs.fedoraproject.org",
    namespace="rpms",
)
CENTOS_DG = DistGitInstance(
    hostname="gitlab.com",
    alternative_hostname=None,
    namespace="redhat/centos-stream/rpms",
)
CBS_CLOUD_DG = DistGitInstance(
    hostname="gitlab.com",
    alternative_hostname=None,
    namespace="CentOS/cloud/rpms",
)


@pytest.mark.parametrize(
    "dg, url, expected",
    (
        pytest.param(FEDORA_DG, "https://src.fedoraproject.org/rpms/packit", True),
        pytest.param(FEDORA_DG, "https://src.fedoraproject.org/rpms/packit.git", True),
        pytest.param(
            FEDORA_DG,
            "ssh://mfocko@pkgs.fedoraproject.org:rpms/packit.git",
            True,
        ),
        pytest.param(FEDORA_DG, "mfocko@pkgs.fedoraproject.org:rpms/packit.git", True),
        pytest.param(FEDORA_DG, "mfocko@pkgs.fedoraproject.org:XXX/packit.git", False),
        pytest.param(CENTOS_DG, "gitlab.com/packit-service/hello-world.git", False),
        pytest.param(
            CENTOS_DG,
            "gitlab.com/packit-service/rpms/hello-world.git",
            False,
        ),
        pytest.param(
            CENTOS_DG,
            "gitlab.com/redhat/centos-stream/rpms/hello-world.git",
            True,
        ),
        pytest.param(
            CBS_CLOUD_DG,
            "gitlab.com/CentOS/cloud/rpms/hello-world.git",
            True,
        ),
    ),
)
def test_has_repository(dg: DistGitInstance, url: str, expected: bool):
    assert dg.has_repository(url) == expected


@pytest.mark.parametrize(
    "url, namespace, expected",
    (
        pytest.param(
            "https://src.fedoraproject.org/",
            "rpms",
            FEDORA_DG,
            id="packit-prod / fedora-source-git-prod",
        ),
        pytest.param(
            "https://gitlab.com/",
            "redhat/centos-stream/rpms",
            CENTOS_DG,
            id="stream-prod",
        ),
        pytest.param(
            "https://gitlab.com/",
            "CentOS/cloud/rpms",
            CBS_CLOUD_DG,
            id="cbs-cloud-sig",
        ),
    ),
)
def test_from_url_and_namespace(url: str, namespace: str, expected: DistGitInstance):
    parsed_dg_instance = DistGitInstance.from_url_and_namespace(url, namespace)

    assert parsed_dg_instance and parsed_dg_instance.hostname in (
        expected.hostname,
        expected.alternative_hostname,
    )
    assert parsed_dg_instance.namespace == expected.namespace


@pytest.mark.parametrize(
    "dg_instance, package, expected_url",
    (
        (
            DISTGIT_INSTANCES["fedpkg"],
            "packit",
            "https://src.fedoraproject.org/rpms/packit",
        ),
        (
            DISTGIT_INSTANCES["centpkg"],
            "nispor",
            "https://gitlab.com/redhat/centos-stream/rpms/nispor",
        ),
        (
            DISTGIT_INSTANCES["fedpkg-stage"],
            "packit",
            "https://src.stg.fedoraproject.org/rpms/packit",
        ),
        (
            DISTGIT_INSTANCES["centpkg-sig"].for_sig("cloud"),
            "hello-world",
            "https://gitlab.com/CentOS/cloud/rpms/hello-world",
        ),
    ),
)
def test_project_url_from_dg_instance(dg_instance, package, expected_url):
    assert dg_instance.distgit_project_url(package=package) == expected_url



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
from pathlib import Path
from typing import Optional

import click
import git

from packit.local_project import CALCULATE, LocalProject, LocalProjectBuilder
from packit.utils.repo import git_remote_url_to_https_url

logger = logging.getLogger(__name__)


class LocalProjectParameter(click.ParamType):
    """
    Path or url.
    """

    name = "path_or_url"

    def __init__(
        self,
        ref_param_name: Optional[str] = None,
        pr_id_param_name: Optional[str] = None,
        merge_pr_param_name: Optional[str] = None,
        target_branch_param_name: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.ref_param_name = ref_param_name
        self.pr_id_param_name = pr_id_param_name
        self.merge_pr_param_name = merge_pr_param_name
        self.target_branch_param_name = target_branch_param_name

    @staticmethod
    def get_param(param_name, ctx):
        value = None
        if param_name in ctx.params:
            value = ctx.params[param_name]
        else:  # use the default
            for param in ctx.command.params:
                if param.name == param_name:
                    value = param.default
        return value

    def convert(self, value, param, ctx):
        if isinstance(value, LocalProject):
            return value

        try:
            pr_id = None
            merge_pr = True
            target_branch = None

            ref = (
                self.get_param(self.ref_param_name, ctx) if self.ref_param_name else ""
            )
            if self.pr_id_param_name:
                pr_id = self.get_param(self.pr_id_param_name, ctx)

            if self.merge_pr_param_name:
                merge_pr = self.get_param(self.merge_pr_param_name, ctx)

            if self.target_branch_param_name:
                target_branch = self.get_param(self.target_branch_param_name, ctx)

            path = Path(value)
            if path.is_file():
                logger.debug(f"Input is file {path}, taking its parent dir.")
                path = path.parent
            # General use-case, create fully featured project
            builder = LocalProjectBuilder()
            if path.is_dir():
                logger.debug(f"Input is a directory: {path.absolute()}")
                local_project = builder.build(
                    working_dir=path.absolute(),
                    ref=ref,
                    remote=ctx.obj.upstream_git_remote,
                    merge_pr=merge_pr,
                    target_branch=target_branch,
                    git_repo=CALCULATE,
                    git_project=CALCULATE,
                    git_service=CALCULATE,
                    full_name=CALCULATE,
                    namespace=CALCULATE,
                    repo_name=CALCULATE,
                    git_url=CALCULATE,
                )
            elif git_remote_url_to_https_url(value):
                logger.debug(f"Input is a URL to a git repo: {value}")
                working_dir = (
                    clone_path
                    if (clone_path := self.get_param("clone_path", ctx))
                    else CALCULATE
                )
                local_project = builder.build(
                    git_url=value,
                    ref=ref,
                    remote=ctx.obj.upstream_git_remote,
                    pr_id=pr_id,
                    merge_pr=merge_pr,
                    target_branch=target_branch,
                    git_repo=CALCULATE,
                    working_dir=working_dir,
                    git_project=CALCULATE,
                    git_service=CALCULATE,
                    full_name=CALCULATE,
                    namespace=CALCULATE,
                    repo_name=CALCULATE,
                )
            else:
                self.fail(
                    "Provided input path_or_url is not a directory nor an URL of a git repo.",
                )

            if not (local_project.working_dir or local_project.git_url):
                self.fail(
                    "Parameter is not an existing directory nor correct git url.",
                    param,
                    ctx,
                )
            return local_project
        except Exception as ex:
            self.fail(ex, param, ctx)


class GitRepoParameter(click.ParamType):
    """Parameter type to represent a Git repository on the local disk, and an
    optional branch, in the format <path>:<branch>.

    Attributes:
        from_ref_param: Name of the CLI parameter which tells the start point of the branch
            to be created, if the branch doesn't exist yet.
    """

    name = "git_repo"

    def __init__(self, from_ref_param: Optional[str] = None):
        super().__init__()
        self.from_ref_param = from_ref_param

    def convert(self, value, param, ctx) -> git.Repo:
        if isinstance(value, git.Repo):
            return value
        if not isinstance(value, str):
            self.fail(f"{value!r} is not a string")

        try:
            path, _, branch = value.partition(":")
            repo = git.Repo(path)
            if not branch:
                return repo

            branch_exists = True
            try:
                repo.rev_parse(branch)
            except git.BadName:
                branch_exists = False

            if self.from_ref_param is not None:
                if ctx.params.get(self.from_ref_param):
                    repo.git.checkout("-B", branch, ctx.params[self.from_ref_param])
                else:
                    self.fail(
                        f"Unable to create branch {branch!r} because "
                        f"{self.from_ref_param!r} is not specified",
                        param,
                        ctx,
                    )
            elif branch_exists:
                repo.git.checkout(branch)
            else:
                self.fail(
                    f"Cannot check out branch {branch!r} because it does not exist",
                    param,
                    ctx,
                )

            return repo

        except git.NoSuchPathError:
            self.fail(f"{path!r} does not exist", param, ctx)
        except git.GitCommandError as ex:
            self.fail(ex, param, ctx)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from packit.utils.commands import cwd, run_command, run_command_remote
from packit.utils.extensions import assert_existence, nested_get
from packit.utils.logging import (
    PackitFormatter,
    StreamLogger,
    commits_to_nice_str,
    set_logging,
)
from packit.utils.repo import (
    commit_message_file,
    get_default_branch,
    get_file_author,
    get_namespace_and_repo_name,
    get_repo,
    git_remote_url_to_https_url,
    is_a_git_ref,
    is_git_repo,
)

__all__ = [
    assert_existence.__name__,
    commit_message_file.__name__,
    commits_to_nice_str.__name__,
    cwd.__name__,
    get_default_branch.__name__,
    get_file_author.__name__,
    get_namespace_and_repo_name.__name__,
    get_repo.__name__,
    git_remote_url_to_https_url.__name__,
    is_a_git_ref.__name__,
    is_git_repo.__name__,
    nested_get.__name__,
    PackitFormatter.__name__,
    run_command.__name__,
    run_command_remote.__name__,
    set_logging.__name__,
    StreamLogger.__name__,
]
OFFENDERS = "!@#$%&*()={[}]|\\'\":;<,>/?`"


def sanitize_branch_name(branch_name: str) -> str:
    """
    replace potentially problematic characters in provided string, a branch name

    e.g. copr says:
        Name must contain only letters, digits, underscores, dashes and dots.
    """
    # https://stackoverflow.com/questions/3411771/best-way-to-replace-multiple-characters-in-a-string
    offenders = OFFENDERS + "^~+"
    for o in offenders:
        branch_name = branch_name.replace(o, "-")
    return branch_name


def sanitize_version(version: str) -> str:
    """Sanitize given string to be usable as a version.

    rpm is picky about release: hates "/" - it's an error
    also prints a warning for "-"

    Follows rules in
    https://github.com/rpm-software-management/rpm/blob/master/docs/manual/spec.md#version

    The version string consists of alphanumeric characters,
    which can optionally be segmented with the separators
    ., _ and +, plus ~ and ^
    """
    for o in OFFENDERS:
        version = version.replace(o, "")
    return version.replace("-", ".")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
E2E tests which utilize cockpit projects
"""
import shutil
from pathlib import Path

import pytest
from flexmock import flexmock

from packit.api import PackitAPI
from packit.config import get_local_package_config
from packit.distgit import DistGit
from packit.local_project import CALCULATE, LocalProjectBuilder
from packit.pkgtool import PkgTool
from packit.utils import repo
from packit.utils.commands import cwd
from tests.spellbook import UP_COCKPIT_OSTREE, get_test_config, initiate_git_repo


@pytest.fixture()
def cockpit_ostree(tmp_path, upstream_without_config):
    u = tmp_path / "up"
    initiate_git_repo(u, tag="179", copy_from=UP_COCKPIT_OSTREE)

    flexmock(repo, get_namespace_and_repo_name=lambda url: ("asd", "qwe"))
    d = tmp_path / "dg"
    d.mkdir()
    initiate_git_repo(d, upstream_remote=upstream_without_config, push=True)

    shutil.copy2(
        UP_COCKPIT_OSTREE / "cockpit-ostree.spec.dg",
        d / "cockpit-ostree.spec",
    )

    return u, d


def test_update_on_cockpit_ostree(cockpit_ostree):
    upstream_path, dist_git_path = cockpit_ostree

    def mocked_new_sources(sources=None):
        sources = sources or []
        if not all(Path(s).is_file() for s in sources):
            raise RuntimeError("archive does not exist")

    flexmock(PkgTool, new_sources=mocked_new_sources)
    flexmock(PackitAPI, init_kerberos_ticket=lambda: None)

    flexmock(
        DistGit,
        push_to_fork=lambda *args, **kwargs: None,
        is_archive_in_lookaside_cache=lambda archive_path: False,
        upload_to_lookaside_cache=lambda archives, pkg_tool, offline: None,
        download_upstream_archives=lambda: ["the-archive"],
    )
    flexmock(DistGit).should_receive("existing_pr").and_return(None)
    flexmock(
        PackitAPI,
        push_and_create_pr=lambda pr_title, pr_description, git_branch, repo, sync_acls: None,
    )

    pc = get_local_package_config(str(upstream_path))
    up_lp = LocalProjectBuilder().build(working_dir=upstream_path, git_repo=CALCULATE)
    c = get_test_config()
    api = PackitAPI(c, pc, up_lp)
    api._dg = DistGit(c, pc)
    api._dg._local_project = LocalProjectBuilder().build(
        working_dir=dist_git_path,
        git_repo=CALCULATE,
    )
    flexmock(api.up, get_specfile_version=lambda: "178")

    with cwd(upstream_path):
        api.sync_release(
            dist_git_branch="main",
            use_local_content=False,
            versions=["179"],
            force_new_sources=False,
            create_pr=True,
        )


def test_update_on_cockpit_ostree_pr_exists(cockpit_ostree):
    upstream_path, dist_git_path = cockpit_ostree

    def mocked_new_sources(sources=None):
        sources = sources or []
        if not all(Path(s).is_file() for s in sources):
            raise RuntimeError("archive does not exist")

    flexmock(PkgTool, new_sources=mocked_new_sources)
    flexmock(PackitAPI, init_kerberos_ticket=lambda: None)

    flexmock(
        DistGit,
        push_to_fork=lambda *args, **kwargs: None,
        is_archive_in_lookaside_cache=lambda archive_path: False,
        upload_to_lookaside_cache=lambda archives, pkg_tool, offline: None,
        download_upstream_archives=lambda: ["the-archive"],
    )
    pr = flexmock(url="https://example.com/pull/1")
    pr.should_receive("update_info").and_return()
    flexmock(DistGit).should_receive("existing_pr").and_return(pr)

    pc = get_local_package_config(str(upstream_path))
    up_lp = LocalProjectBuilder().build(working_dir=upstream_path, git_repo=CALCULATE)
    c = get_test_config()
    api = PackitAPI(c, pc, up_lp)
    api._dg = DistGit(c, pc)
    api._dg._local_project = LocalProjectBuilder().build(
        working_dir=dist_git_path,
        git_repo=CALCULATE,
    )
    flexmock(api.up, get_specfile_version=lambda: "178")

    with cwd(upstream_path):
        assert pr == api.sync_release(
            dist_git_branch="main",
            use_local_content=False,
            versions=["179"],
            force_new_sources=False,
            create_pr=True,
        )


def test_srpm_on_cockpit_ostree(cockpit_ostree):
    upstream_path, dist_git_path = cockpit_ostree

    pc = get_local_package_config(str(upstream_path))
    up_lp = LocalProjectBuilder().build(working_dir=upstream_path, git_repo=CALCULATE)
    c = get_test_config()
    api = PackitAPI(c, pc, up_lp)

    with cwd(upstream_path):
        api.create_srpm()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import subprocess
from pathlib import Path

import git
import pytest
from specfile import Specfile
from specfile.exceptions import SourceNumberException

from packit.constants import DISTRO_DIR
from packit.exceptions import PackitException
from packit.patches import PatchGenerator
from packit.utils.commands import cwd
from tests.integration.conftest import mock_spec_download_remote_s
from tests.spellbook import (
    TARBALL_NAME,
    build_srpm,
    create_git_am_style_history,
    create_history_with_empty_commit,
    create_history_with_patch_ids,
    create_merge_commit_in_source_git,
    create_patch_mixed_history,
    git_add_and_commit,
    run_prep_for_srpm,
)


def test_update_dist_git_with_sync_status_check(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_update_source_git,
):
    """Check that exception is raised when there's a commit
    in dist-git which has no origin in source-git.
    """

    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote

    api_instance_update_source_git.dg.specfile.version = "0.2.0"
    api_instance_update_source_git.dg.commit("Extra dist-git commit", "")

    api_instance_update_source_git.up.specfile.version = "0.3.0"
    api_instance_update_source_git.up.commit("Source-git commit to be synced", "")

    with pytest.raises(PackitException) as exc:
        api_instance_update_source_git.update_dist_git(
            version=None,
            upstream_ref=None,
            add_new_sources=False,
            force_new_sources=False,
            upstream_tag=None,
            commit_title="",
            commit_msg="",
            check_sync_status=True,
        )
    assert " have diverged." in str(exc.value)


def test_update_dist_git_dist_git_not_pristine(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_update_source_git,
):
    """Check that exception is raised when the
    dist-git repo is not pristine.
    """

    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote

    (distgit / "new_file").write_text("abcd")

    api_instance_update_source_git.up.specfile.version = "0.3.0"
    api_instance_update_source_git.up.commit("Source-git commit to be synced", "")

    with pytest.raises(PackitException) as exc:
        api_instance_update_source_git.update_dist_git(
            version=None,
            upstream_ref=None,
            add_new_sources=False,
            force_new_sources=False,
            upstream_tag=None,
            commit_title="",
            commit_msg="",
            check_sync_status=True,
        )
    assert "is not pristine" in str(exc.value)


def test_basic_local_update_without_patching(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
):
    """propose-downstream for sourcegit test: mock remote API, use local upstream and dist-git
    Check that the upstream commit hash is saved when 'mock_commit_origin' is set.
    """

    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    mock_spec_download_remote_s(distgit)

    api_instance_source_git.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
        upstream_ref="0.1.0",
        mark_commit_origin=True,
    )

    assert (distgit / TARBALL_NAME).is_file()
    spec = Specfile(distgit / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    assert (
        f"From-source-git-commit: {git.Repo(sourcegit).head.commit.hexsha}"
        in git.Repo(distgit).head.commit.message
    )


@pytest.mark.parametrize("ref", [None, "0.1.0", "0.1*", "0.*"])
def test_basic_local_update_empty_patch(
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
    ref,
):
    """propose-downstream for sourcegit test: mock remote API, use local upstream and dist-git
    Check that by default commit origin is not marked in dist-git.
    """

    distgit, _ = distgit_and_remote
    mock_spec_download_remote_s(distgit)
    api_instance_source_git.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
        upstream_ref=ref,
    )

    assert (distgit / TARBALL_NAME).is_file()
    spec = Specfile(distgit / "beer.spec")
    assert spec.expanded_version == "0.1.0"

    with spec.patches() as patches:
        assert not patches
    assert "From-source-git-commit" not in git.Repo(distgit).head.commit.message


def test_basic_local_update_patch_content(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
):
    """propose-downstream for sourcegit test: mock remote API, use local upstream and dist-git
    Check that commit origin is not marked when 'mark_commit_origin' is set to False.
    """

    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    mock_spec_download_remote_s(distgit)

    create_merge_commit_in_source_git(sourcegit)

    source_file = sourcegit / "big-source-file.txt"
    source_file.write_text("new changes")
    git_add_and_commit(directory=sourcegit, message="source change")

    source_file = sourcegit / "ignored_file.txt"
    source_file.write_text(" And I am sad.")
    git_add_and_commit(directory=sourcegit, message="make a file sad")

    api_instance_source_git.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
        upstream_ref="0.1.0",
        mark_commit_origin=False,
    )

    git_diff = subprocess.check_output(
        ["git", "diff", "HEAD~", "HEAD"],
        cwd=distgit,
    ).decode()

    assert "From-source-git-commit" not in git.Repo(distgit).head.commit.message
    assert (
        """
-Version:        0.0.0
+Version:        0.1.0"""
        in git_diff
    )
    # Make sure the patches are placed after Source0, but outside %if %endif
    patches = """\
Source0:        %{upstream_name}-%{version}.tar.gz
 %endif
+# switching to amarillo hops
+# Author: Packit Test Suite <test@example.com>
+Patch0001:      0001-switching-to-amarillo-hops.patch
+# actually, let's do citra
+# Author: Packit Test Suite <test@example.com>
+Patch0002:      0002-actually-let-s-do-citra.patch
+# source change
+# Author: Packit Test Suite <test@example.com>
+Patch0003:      0003-source-change.patch
 BuildArch:      noarch
"""
    assert patches in git_diff

    assert "Patch0004:" not in git_diff

    assert (
        """ - 0.1.0-1
+- Initial brewing
+
 * Sun Feb 24 2019 Tomas Tomecek <ttomecek@redhat.com> - 0.0.0-1
 - No brewing, yet."""
        in git_diff
    )

    patch_1_3 = """
+Subject: [PATCH 1/3] switching to amarillo hops
+
+---
+ hops | 2 +-
+ 1 file changed, 1 insertion(+), 1 deletion(-)
+
+diff --git a/hops b/hops"""
    assert patch_1_3 in git_diff
    assert (
        """\
+--- a/hops
++++ b/hops
+@@ -1 +1 @@
+-Cascade
++Amarillo
+--"""
        in git_diff
    )

    assert (
        """\
+Subject: [PATCH 2/3] actually, let's do citra
+
+---
+ hops | 2 +-
+ 1 file changed, 1 insertion(+), 1 deletion(-)
+
+diff --git a/hops b/hops"""
        in git_diff
    )
    assert (
        (
            """\
+--- a/hops
++++ b/hops
+@@ -1 +1 @@
+-Amarillo
++Citra
+--"""
        )
        in git_diff
    )

    assert (
        """
+--- a/big-source-file.txt
++++ b/big-source-file.txt
+@@ -1,2 +1 @@
+-This is a testing file
+-containing some text.
++new changes"""
        in git_diff
    )

    # diff of the source files (not synced) should not be directly in the git diff
    assert (
        """
+Subject: [PATCH 3/3] source change
+
+---
+ big-source-file.txt | 3 +--
+ 1 file changed, 1 insertion(+), 2 deletions(-)
+
+diff --git a/big-source-file.txt b/big-source-file.txt"""
        in git_diff
    )

    # ignored file should not be in the diff
    assert "--- a/ignored_file.txt\n" not in git_diff


def test_basic_local_update_patch_content_with_metadata(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
):
    """propose-downstream for sourcegit test: mock remote API, use local upstream and dist-git"""

    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    mock_spec_download_remote_s(distgit)

    create_merge_commit_in_source_git(sourcegit)

    source_file = sourcegit / "big-source-file.txt"
    source_file.write_text("new changes")
    git_add_and_commit(
        directory=sourcegit,
        message="source change\n"
        "patch_name: testing.patch\n"
        "description: Few words for info.",
    )

    source_file = sourcegit / "ignored_file.txt"
    source_file.write_text(" And I am sad.")
    git_add_and_commit(directory=sourcegit, message="make a file sad")

    api_instance_source_git.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
        upstream_ref="0.1.0",
    )

    git_diff = subprocess.check_output(
        ["git", "diff", "HEAD~", "HEAD"],
        cwd=distgit,
    ).decode()

    patches = """
+# switching to amarillo hops
+# Author: Packit Test Suite <test@example.com>
+Patch0001:      0001-switching-to-amarillo-hops.patch
+# actually, let's do citra
+# Author: Packit Test Suite <test@example.com>
+Patch0002:      0002-actually-let-s-do-citra.patch
+# Few words for info.
+Patch0003:      testing.patch
"""
    assert patches in git_diff


def test_basic_local_update_patch_content_with_metadata_and_patch_ignored(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
):
    """propose-downstream for sourcegit test: mock remote API, use local upstream and dist-git"""

    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    mock_spec_download_remote_s(distgit)

    create_merge_commit_in_source_git(sourcegit)

    source_file = sourcegit / "big-source-file.txt"
    source_file.write_text("new changes")
    git_add_and_commit(
        directory=sourcegit,
        message="source change\nignore: true",
    )

    source_file = sourcegit / "ignored_file.txt"
    source_file.write_text(" And I am sad.")
    git_add_and_commit(directory=sourcegit, message="make a file sad")

    api_instance_source_git.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
        upstream_ref="0.1.0",
    )

    git_diff = subprocess.check_output(
        ["git", "diff", "HEAD~", "HEAD"],
        cwd=distgit,
    ).decode()

    patches = """
+# switching to amarillo hops
+# Author: Packit Test Suite <test@example.com>
+Patch0001:      0001-switching-to-amarillo-hops.patch
+# actually, let's do citra
+# Author: Packit Test Suite <test@example.com>
+Patch0002:      0002-actually-let-s-do-citra.patch
"""
    assert patches in git_diff


def test_basic_local_update_patch_content_with_downstream_patch(
    sourcegit_and_remote,
    distgit_and_remote,
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
):
    """propose-downstream for sourcegit test: mock remote API, use local upstream and dist-git"""

    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    mock_spec_download_remote_s(distgit)

    create_merge_commit_in_source_git(sourcegit)

    source_file = sourcegit / "ignored_file.txt"
    source_file.write_text(" And I am sad.")
    git_add_and_commit(directory=sourcegit, message="make a file sad")

    api_instance_source_git.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
        upstream_ref="0.1.0",
    )

    git_diff = subprocess.check_output(
        ["git", "diff", "HEAD~", "HEAD"],
        cwd=distgit,
    ).decode()

    patches = """
+# switching to amarillo hops
+# Author: Packit Test Suite <test@example.com>
+Patch0001:      0001-switching-to-amarillo-hops.patch
+# actually, let's do citra
+# Author: Packit Test Suite <test@example.com>
+Patch0002:      0002-actually-let-s-do-citra.patch
"""
    assert patches in git_diff


@pytest.mark.parametrize("ref", ["0.1.0", "0.1*", "0.*"])
def test_srpm(mock_remote_functionality_sourcegit, api_instance_source_git, ref):
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, "0.1.0")
    create_merge_commit_in_source_git(sg_path)
    with cwd(sg_path):
        api_instance_source_git.create_srpm(upstream_ref=ref)
    srpm_path = next(sg_path.glob("beer-0.1.0-2.*.src.rpm"))
    assert srpm_path.is_file()
    build_srpm(srpm_path)
    branches = subprocess.check_output(
        ["git", "for-each-ref", "--format=%(refname:short)", "refs/heads/"],
        cwd=sg_path,
    ).split(b"\n")
    for b in branches:
        if b and b.startswith(b"packit-patches-"):
            raise AssertionError(
                "packit-patches- branch was found - the history shouldn't have been linearized",
            )
    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "0001-switching-to-amarillo-hops.patch",
        "0002-actually-let-s-do-citra.patch",
    }


@pytest.mark.parametrize("ref", ["0.1.0", "0.1*", "0.*"])
def test_srpm_merge_storm(
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
    ref,
):
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, "0.1.0")
    create_merge_commit_in_source_git(sg_path, go_nuts=True)

    # linearization creates a new branch, make some arbitrary moves to verify
    # we end up in the former branch after the build
    subprocess.check_call(["git", "checkout", "-B", "test-branch"], cwd=sg_path)
    subprocess.check_call(["git", "checkout", "main"], cwd=sg_path)

    with cwd(sg_path):
        api_instance_source_git.create_srpm(upstream_ref=ref)
    srpm_path = next(sg_path.glob("beer-0.1.0-2.*.src.rpm"))
    assert srpm_path.is_file()
    build_srpm(srpm_path)
    branches = subprocess.check_output(
        ["git", "for-each-ref", "--format=%(refname:short)", "refs/heads/"],
        cwd=sg_path,
    ).split(b"\n")
    for b in branches:
        if b and b.startswith(b"packit-patches-"):
            break
    else:
        raise AssertionError(
            "packit-patches- branch was not found - this should trigger the linearization",
        )
    # make sure we are on the main branch
    assert (
        subprocess.check_output(["git", "branch", "--show-current"], cwd=sg_path)
        .decode()
        .strip()
        == "main"
    )
    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "0001-MERGE-COMMIT.patch",
        "0002-ugly-merge-commit.patch",
    }


def test_srpm_merge_storm_dirty(api_instance_source_git):
    """verify the linearization is halted when a source-git repo si dirty"""
    ref = "0.1.0"
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, ref)
    create_merge_commit_in_source_git(sg_path, go_nuts=True)
    (sg_path / "malt").write_text("Mordor\n")
    with pytest.raises(PackitException) as ex, cwd("/"):
        # let's mimic p-s by having different cwd than the project
        api_instance_source_git.create_srpm(upstream_ref=ref)
    assert "The source-git repo is dirty" in str(ex.value)


def test_linearization(api_instance_source_git):
    ref = "0.1.0"
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, ref)
    create_merge_commit_in_source_git(sg_path, go_nuts=True)
    with cwd("/"):  # let's mimic p-s by having different cwd than the project
        pg = PatchGenerator(api_instance_source_git.upstream_local_project)
        pg.create_patches(ref, sg_path / DISTRO_DIR)
    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "0001-sourcegit-content.patch",
        "0002-MERGE-COMMIT.patch",
        "0003-ugly-merge-commit.patch",
    }


@pytest.mark.parametrize("ref", ["0.1.0", "0.1*", "0.*"])
def test_srpm_git_am(mock_remote_functionality_sourcegit, api_instance_source_git, ref):
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, "0.1.0")

    with api_instance_source_git.up.specfile.prep() as prep:
        prep.autosetup.options.S = "git_am"

    create_git_am_style_history(sg_path)

    with cwd(sg_path):
        api_instance_source_git.create_srpm(upstream_ref=ref)

    srpm_path = next(sg_path.glob("beer-0.1.0-2.*.src.rpm"))
    assert srpm_path.is_file()
    build_srpm(srpm_path)

    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "citra.patch",
        "0001-m04r-malt.patch",
        "malt.patch",
    }
    run_prep_for_srpm(srpm_path)
    if (sg_path / "beer-0.1.0-build").is_dir():
        prep_root = sg_path / "beer-0.1.0-build" / "beerware-0.1.0"
    else:
        prep_root = sg_path / "beerware-0.1.0"
    assert prep_root.joinpath("malt").read_text() == "Munich\n"
    assert prep_root.joinpath("hops").read_text() == "Saaz\n"


@pytest.mark.parametrize("ref", ["0.1.0", "0.1*", "0.*"])
def test_srpm_git_no_prefix_patches(
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
    ref,
):
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, "0.1.0")

    with api_instance_source_git.up.specfile.sections() as sections:
        sections.package[10:10] = (
            "Patch1: amarillo.patch",
            "Patch2: citra.patch",
            "Patch8: malt.patch",
        )
        sections.prep[0:2] = [
            "%setup -n %{upstream_name}-%{version}",
            "%patch -P1 -p1",
            "%patch -P2 -p0",
            "%patch -P8 -p1",
        ]

    create_patch_mixed_history(sg_path)

    with cwd(sg_path):
        api_instance_source_git.create_srpm(upstream_ref=ref)

    srpm_path = next(sg_path.glob("beer-0.1.0-2.*.src.rpm"))
    assert srpm_path.is_file()
    build_srpm(srpm_path)

    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "amarillo.patch",
        "citra.patch",
        "malt.patch",
    }


@pytest.mark.parametrize("ref", ["0.1.0", "0.1*", "0.*"])
def test_srpm_empty_patch(
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
    ref,
):
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, "0.1.0")

    with api_instance_source_git.up.specfile.sections() as sections:
        sections.package[10:10] = (
            "Patch1: amarillo.patch",
            "Patch2: citra.patch",
            "Patch5: saaz.patch",
            "Patch8: malt.patch",
        )

    create_history_with_empty_commit(sg_path)

    with cwd(sg_path):
        api_instance_source_git.create_srpm(upstream_ref=ref)

    srpm_path = next(sg_path.glob("beer-0.1.0-2.*.src.rpm"))
    assert srpm_path.is_file()
    build_srpm(srpm_path)

    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "amarillo.patch",
        "citra.patch",
        "saaz.patch",
        "malt.patch",
    }
    assert sg_path.joinpath(DISTRO_DIR, "saaz.patch").read_text() == ""


@pytest.mark.parametrize("ref", ["0.1.0", "0.1*", "0.*"])
def test_srpm_patch_non_conseq_indices(
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
    ref,
):
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, "0.1.0")

    with api_instance_source_git.up.specfile.sections() as sections:
        sections.package[10:10] = (
            "Patch0: amarillo.patch",
            "Patch3: citra.patch",
            "Patch4: saaz.patch",
            "Patch5: malt.patch",
        )

    create_history_with_empty_commit(sg_path)

    malt = sg_path.joinpath("malt")
    malt.write_text("Wheat\n")
    git_add_and_commit(directory=sg_path, message="Weißbier! Summer is coming!")

    with cwd(sg_path):
        api_instance_source_git.create_srpm(upstream_ref=ref)

    # make sure the patch is inserted AFTER existing patches
    with api_instance_source_git.up.specfile.patches() as patches:
        last_patch = patches[-1]
    assert last_patch.number == 6
    assert last_patch.filename == "0004-Wei-bier-Summer-is-coming.patch"

    srpm_path = next(sg_path.glob("beer-0.1.0-2.*.src.rpm"))
    assert srpm_path.is_file()
    build_srpm(srpm_path)

    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "amarillo.patch",
        "citra.patch",
        "saaz.patch",
        "malt.patch",
        "0004-Wei-bier-Summer-is-coming.patch",
    }
    assert sg_path.joinpath(DISTRO_DIR, "saaz.patch").read_text() == ""


@pytest.mark.parametrize("starting_patch_id", [0, 1, 100])
def test_add_patch_with_patch_id(api_instance_source_git, starting_patch_id):
    """check that patches with patch_id set are added to spec correctly"""
    spec_dir = api_instance_source_git.up.absolute_specfile_dir
    spec: Specfile = api_instance_source_git.up.specfile

    # we want to add this patch to the spec
    good_patch_name1 = "hello.patch"
    # we need to create the patch file so that specfile can find it and process it
    good_patch_path1 = spec_dir.joinpath(good_patch_name1)
    good_patch_path1.write_text("")
    spec.add_patch(good_patch_name1, starting_patch_id)

    with spec.patches() as patches:
        assert patches[0].number == starting_patch_id

    # add another, this time without patch_id
    good_patch_name2 = "hello2.patch"
    good_patch_path2 = spec_dir.joinpath(good_patch_name2)
    good_patch_path2.write_text("")
    spec.add_patch(good_patch_name2)

    # check that index of the second patch is (starting + 1)
    with spec.patches() as patches:
        assert patches[1].number == starting_patch_id + 1

    # and now another with an index lower or equal than the last one and check if
    # an exc is thrown b/c that's not supported
    # to change order of patches (people should reorder the git history instead)
    patch_name = "nope.patch"
    bad_patch_id = (
        starting_patch_id + 1 if starting_patch_id <= 1 else starting_patch_id - 1
    )
    with pytest.raises(SourceNumberException):
        spec.add_patch(patch_name, bad_patch_id)


def test_add_patch_first_id_1(api_instance_source_git):
    """check that add_patch sets the first patch id to 1"""
    spec_dir = api_instance_source_git.up.absolute_specfile_dir
    spec: Specfile = api_instance_source_git.up.specfile

    # we want to add this patch to the spec
    good_patch_name1 = "hello.patch"
    # we need to create the patch file so that specfile can find it and process it
    good_patch_path1 = spec_dir.joinpath(good_patch_name1)
    good_patch_path1.write_text("")
    spec.add_patch(good_patch_name1, initial_number=1)

    with spec.patches() as patches:
        assert patches[0].number == 1


def test_srpm_add_patch_with_ids(
    mock_remote_functionality_sourcegit,
    api_instance_source_git,
):
    ref = "0.1.0"
    sg_path = Path(api_instance_source_git.upstream_local_project.working_dir)
    mock_spec_download_remote_s(sg_path, sg_path / DISTRO_DIR, ref)

    create_history_with_patch_ids(sg_path)

    with cwd(sg_path):
        api_instance_source_git.create_srpm(upstream_ref=ref)

    srpm_path = next(sg_path.glob("beer-0.1.0-2.*.src.rpm"))
    assert srpm_path.is_file()
    build_srpm(srpm_path)

    assert {x.name for x in sg_path.joinpath(DISTRO_DIR).glob("*.patch")} == {
        "amarillo.patch",
        "citra.patch",
        "malt.patch",
    }
    with api_instance_source_git.up.specfile.patches() as patches:
        assert patches[0].filename == "amarillo.patch"
        assert patches[0].number == 3
        assert patches[1].filename == "citra.patch"
        assert patches[1].number == 4
        assert patches[2].filename == "malt.patch"
        assert patches[2].number == 100



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from flexmock import flexmock
from git import PushInfo

from packit.actions import ActionName
from packit.base_git import PackitRepositoryBase
from packit.config import CommonPackageConfig, Config, PackageConfig, RunCommandType
from packit.exceptions import PackitException
from packit.local_project import LocalProject, LocalProjectBuilder
from tests.spellbook import can_a_module_be_imported


@pytest.mark.parametrize(
    "echo_cmd, expected_output",
    [
        ("echo 'hello world'", ["hello world\n"]),
        # should return output of only the last one
        (["echo 'ignore me'", "echo 'hello world'"], ["ignore me\n", "hello world\n"]),
    ],
)
def test_get_output_from_action_defined(echo_cmd, expected_output):
    packit_repository_base = PackitRepositoryBase(
        config=flexmock(Config()),
        package_config=flexmock(
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        actions={ActionName.pre_sync: echo_cmd},
                    ),
                },
            ),
        ),
    )

    packit_repository_base.local_project = flexmock(working_dir=".")

    result = packit_repository_base.actions_handler.get_output_from_action(
        ActionName.pre_sync,
    )
    assert result == expected_output


@pytest.mark.skipif(
    not can_a_module_be_imported("sandcastle"),
    reason="sandcastle is not installed",
)
def test_get_output_from_action_defined_in_sandcastle():
    from sandcastle.api import Sandcastle

    echo_cmd = "hello world"
    flexmock(Sandcastle).should_receive("get_api_client")
    flexmock(Sandcastle).should_receive("is_pod_already_deployed").and_return(True)
    c = Config()
    c.command_handler = RunCommandType.sandcastle
    packit_repository_base = PackitRepositoryBase(
        config=c,
        package_config=PackageConfig(
            packages={
                "package": CommonPackageConfig(actions={ActionName.pre_sync: echo_cmd}),
            },
        ),
    )
    packit_repository_base.local_project = LocalProjectBuilder().build(
        working_dir="/tmp",
    )

    flexmock(Sandcastle).should_receive("run")
    flexmock(Sandcastle).should_receive("exec").and_return(echo_cmd)
    flexmock(Sandcastle).should_receive("delete_pod").and_return(None)
    result = packit_repository_base.actions_handler.get_output_from_action(
        ActionName.pre_sync,
    )
    assert result[-1] == echo_cmd


@pytest.mark.skip(
    reason="Skipping since we don't have an OpenShift cluster by default.",
)
def test_run_in_sandbox():
    packit_repository_base = PackitRepositoryBase(
        config=Config(),
        package_config=PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    actions={ActionName.pre_sync: "ls -lha"},
                ),
            },
        ),
    )
    packit_repository_base.config.actions_handler = "sandcastle"

    result = packit_repository_base.actions_handler.get_output_from_action(
        ActionName.pre_sync,
    )
    assert "total 4.0K" in result
    assert "drwxr-xr-x. 1 root root" in result


def test_base_push_bad(distgit_and_remote):
    distgit, _ = distgit_and_remote

    b = PackitRepositoryBase(
        config=Config(),
        package_config=PackageConfig(packages={"package": CommonPackageConfig()}),
    )
    b.local_project = LocalProjectBuilder().build(
        working_dir=distgit,
        git_url="https://github.com/packit/lol",
    )
    flexmock(LocalProject).should_receive("git_repo").and_return(
        flexmock()
        .should_receive("remote")
        .and_return(
            flexmock(
                push=lambda *args, **kwargs: [
                    PushInfo(PushInfo.REMOTE_REJECTED, None, None, None, None),
                ],
            ),
        )
        .mock(),
    )
    with pytest.raises(PackitException) as e:
        b.push("master")
    assert "unable to push" in str(e.value)


def test_base_push_good(distgit_and_remote):
    distgit, _ = distgit_and_remote

    b = PackitRepositoryBase(
        config=Config(),
        package_config=PackageConfig(packages={"package": CommonPackageConfig()}),
    )
    b.local_project = LocalProjectBuilder().build(
        working_dir=distgit,
        git_url="https://github.com/packit/lol",
    )
    flexmock(LocalProject).should_receive("git_repo").and_return(
        flexmock()
        .should_receive("remote")
        .and_return(
            flexmock(
                push=lambda *args, **kwargs: [
                    PushInfo(PushInfo.FAST_FORWARD, None, None, None, None),
                ],
            ),
        )
        .mock(),
    )
    b.push("master")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Common package config attributes so they can be imported both in PackageConfig and JobConfig
"""
import logging
import warnings
from enum import Enum
from os import getenv
from os.path import basename
from re import split
from typing import Any, Optional, Union

from packit.actions import ActionName
from packit.config.commands import TestCommandConfig
from packit.config.notifications import (
    NotificationsConfig,
)
from packit.config.requirements import RequirementsConfig
from packit.config.sources import SourcesItem
from packit.constants import DISTGIT_INSTANCES
from packit.dist_git_instance import DistGitInstance
from packit.exceptions import PackitConfigException
from packit.sync import SyncFilesItem, iter_srcs


class Deployment(Enum):
    dev = "dev"
    stg = "stg"
    prod = "prod"


class MockBootstrapSetup(Enum):
    default = "default"
    on = "on"
    off = "off"
    image = "image"


def _construct_dist_git_instance(
    base_url: Optional[str],
    namespace: Optional[str],
    pkg_tool: Optional[str],
    sig: Optional[str] = None,
) -> DistGitInstance:
    """Construct a dist-git instance information from provided configuration.

    Args:
        base_url: Base URL of the dist-git provided from the config.
        namespace: Namespace in the dist-git provided from the config.
        pkg_tool: Packaging tool to be used provided from the config.
        sig: SIG that maintains the “downstream” package. Used for adjusting the
            namespace.

            Defaults to `None`.

    Returns:
        Dist-git instance information that is used in config.
    """

    # explicitly specified values in config override everything
    if base_url is not None:
        return DistGitInstance.from_url_and_namespace(base_url, namespace)

    # explicitly specified packaging tool overrides too
    if pkg_tool:
        return DISTGIT_INSTANCES[pkg_tool].for_sig(sig=sig)

    # we try the environment variables
    base_url, namespace = getenv("DISTGIT_URL"), getenv("DISTGIT_NAMESPACE")
    if base_url is not None:
        return DistGitInstance.from_url_and_namespace(base_url, namespace)

    # if nothing has been provided, default to the Fedora
    return DISTGIT_INSTANCES["fedpkg"]


class CommonPackageConfig:
    """Common configuration

    Attributes:
        config_file_path: Path of the configuration file from which this
            configuration was read.
        specfile_path: Path of the specfile in the upstream repo.
        _files_to_sync: List of files to be synced from the upstream
            repo to dist-git.
        _file_to_sync_used: Whether the _files_to_sync config is used.
            Flag used to provide backwards compatibility with 'synced_files'.
        synced_files: Deprecated. List of files to be synced from the upstream
            repo to dist-git.
        patch_generation_ignore_paths: Paths in the upstream repo to be ignored when
            generating patches.
        patch_generation_patch_id_digits: Number of digits used to produce patch IDs when
            adding 'PatchN' tags to the specfile.
        upstream_project_url: URL of the upstream project.
        upstream_project_name: Name of the upstream project.
        downstream_package_name: Name of the package downstream (i.e. in dist-git)
        paths: List of relative paths in the upstream repository, which should be
            considered for this package.
        _downstream_project_url: Deprecated. URL of the dist-git repo.
        dist_git_base_url: Base URL of the dist-git forge where the downstream package is
            stored.
        dist_git_namespace: Namespace in dist-git where the downstream package
            is stored.
        actions: Custom steps used during different operations.
        upstream_ref: The ref used for the last upstream release.
        allowed_gpg_keys: GPG-key fingerprints allowed to sign commits to be
            proposed to dist-git.
        create_pr: Whether to create a PR when proposing an update to dist-git.
        sync_changelog: Sync the changelog part of the specfile when proposing
            and update to dist-git.
        create_sync_note: Whether to create README.packit when proposing an
            update to dist-git.
        spec_source_id: The 'Source' tag in the specfile to be updated.
        spec_source_id_number: The numeric part of the above
        notifications: Notifications to send on a successful build.
        identifier: Suffix to be added to checks. Used to differentiate check flags
            which otherwise would have identical names.
        packit_instances: Packit-as-a-Service instances to be used.
        upstream_tag_template: Template used to transform version numbers to Git-tags.
        archive_root_dir_template: Template to generate the root directory used
            in source archives.
        copy_upstream_release_description: Use the GitHub release notes to update
            the changelog when proposing updates to dist-git.
        sources: URLs to sources to be used during SRPM builds (source-git specific).
        merge_pr_in_ci: Whether to merge a PR in the main branch prior to building.
        srpm_build_deps: Additional dependencies needed by SRPM builds.
        issue_repository: Repo were issues for downstream job failures (Koji builds,
            Bodhi updates) should be filed.
        release_suffix: String that can be used to override the default release
            suffix generated by Packit.
        update_release: Whether to update Release.
        _targets: copr_build, mock chroots where to build tests, builds to test.
            The _ prefix is used because 'targets' without it is now used for
            backward compatibility.
        timeout: copr_build, give up watching a build after timeout, defaults to 7200s
        owner: copr_build, a namespace in COPR where the build should happen
        project: copr_build, a name of the copr project
        dist_git_branches: propose_downstream, branches in dist-git where packit should work
        branch: for `commit` trigger to specify the branch name
        scratch: if we want to run scratch build in koji
        list_on_homepage: if set, created copr project will be visible on copr's home-page
        preserve_project: if set, project will not be created as temporary
        additional_packages: buildroot packages for the chroot [DOES NOT WORK YET]
        additional_repos: buildroot additional additional_repos
        bootstrap: mock bootstrap feature setup
        fmf_url: - git repository containing the metadata (FMF) tree
        fmf_ref: - branch, tag or commit specifying the desired git revision
        fmf_path: - path to the fmf root
        use_internal_tf: if we want to use internal instance of Testing Farm
        skip_build: if we want to skip build phase for Testing Farm job
        env: environment variables
        enable_net: if set to False, Copr builds have network disabled
        allowed_pr_authors: List of Fedora accounts for which distgit PRs we
            will run koji builds.
        allowed_committers: List of Fedora accounts for which distgit pushes we
            will run koji builds.
        allowed_builders: List of Fedora accounts whose Koji builds will trigger
            Bodhi updates.
        tmt_plan: Test to run in Testing Farm.
        tf_post_install_script: post install script to run before the tf tests
        tf_extra_params: Additional Testing Farm parameters to merge into the
            payload of TF requests.
        module_hotfixes: if set, copr will generate repo files with module_hotfixes=1
        upload_sources: If Packit should upload sources to lookaside cache. True by default.
        pkg_tool: Tool that is used for interaction with the lookaside cache.
        version_update_mask: String containing a reg exp. The old version contained in the
            specfile and the newly released version have both to match this reg exp
            otherwise Packit shall not sync the release downstream.
        parse_time_macros: Dict with macros to (un)define before parsing specfile.
            Keys are macro names and values are macro values. A value of None will undefine
            the corresponding macro.
        status_name_template: Template for configurable names for status checks.
        sync_test_job_statuses_with_builds: Whether to sync test job statuses with corresponding
            build job statuses.
        sig: Special interest group (SIG) that maintains the “downstream” package.
        osh_diff_scan_after_copr_build: Whether to run a differential scan in
            OpenScanHub after successful Copr build.
        csmock_args: Pass additional arguments to csmock
    """

    def __init__(
        self,
        config_file_path: Optional[str] = None,
        specfile_path: Optional[str] = None,
        synced_files: Optional[list[SyncFilesItem]] = None,
        files_to_sync: Optional[list[SyncFilesItem]] = None,
        dist_git_namespace: Optional[str] = None,
        upstream_project_url: Optional[str] = None,  # can be URL or path
        upstream_package_name: Optional[str] = None,
        paths: Optional[list[str]] = None,
        downstream_project_url: Optional[str] = None,
        downstream_package_name: Optional[str] = None,
        dist_git_base_url: Optional[str] = None,
        actions: Optional[dict[ActionName, Union[str, list[str]]]] = None,
        upstream_ref: Optional[str] = None,
        allowed_gpg_keys: Optional[list[str]] = None,
        create_pr: bool = True,
        sync_changelog: bool = False,
        create_sync_note: bool = True,
        spec_source_id: str = "Source0",
        upstream_tag_template: str = "{version}",
        archive_root_dir_template: str = "{upstream_pkg_name}-{version}",
        patch_generation_ignore_paths: Optional[list[str]] = None,
        patch_generation_patch_id_digits: int = 4,
        notifications: Optional[NotificationsConfig] = None,
        copy_upstream_release_description: bool = False,
        sources: Optional[list[SourcesItem]] = None,
        merge_pr_in_ci: bool = True,
        srpm_build_deps: Optional[list[str]] = None,
        identifier: Optional[str] = None,
        packit_instances: Optional[list[Deployment]] = None,
        issue_repository: Optional[str] = None,
        release_suffix: Optional[str] = None,
        update_release: bool = True,
        # Former JobMetadataConfig attributes
        _targets: Union[list[str], dict[str, dict[str, Any]], None] = None,
        timeout: int = 7200,
        owner: Optional[str] = None,
        project: Optional[str] = None,
        dist_git_branches: Union[list[str], dict[str, dict[str, list]], None] = None,
        branch: Optional[str] = None,
        scratch: bool = False,
        list_on_homepage: bool = False,
        preserve_project: bool = False,
        additional_packages: Optional[list[str]] = None,
        additional_repos: Optional[list[str]] = None,
        bootstrap: Optional[MockBootstrapSetup] = None,
        fmf_url: Optional[str] = None,
        fmf_ref: Optional[str] = None,
        fmf_path: Optional[str] = None,
        use_internal_tf: bool = False,
        skip_build: bool = False,
        env: Optional[dict[str, Any]] = None,
        enable_net: bool = False,
        allowed_pr_authors: Optional[list[str]] = None,
        allowed_committers: Optional[list[str]] = None,
        allowed_builders: Optional[list[str]] = None,
        tmt_plan: Optional[str] = None,
        tf_post_install_script: Optional[str] = None,
        tf_extra_params: Optional[dict[Any, Any]] = None,
        module_hotfixes: bool = False,
        # # vm-image-build
        # example: "rhel-86"
        image_distribution: Optional[str] = None,
        image_request: Optional[dict] = None,
        image_customizations: Optional[dict] = None,
        copr_chroot: Optional[str] = None,
        follow_fedora_branching: bool = False,
        upstream_tag_include: str = "",
        upstream_tag_exclude: str = "",
        prerelease_suffix_pattern: str = r"([.\-_~^]?)(alpha|beta|rc|pre(view)?)([.\-_]?\d+)?",
        prerelease_suffix_macro: Optional[str] = None,
        upload_sources: bool = True,
        pkg_tool: Optional[str] = None,
        version_update_mask: Optional[str] = None,
        test_command: Optional[TestCommandConfig] = None,
        parse_time_macros: Optional[dict[str, str]] = None,
        require: Optional[RequirementsConfig] = None,
        status_name_template: Optional[str] = None,
        sync_test_job_statuses_with_builds: bool = True,
        sig: Optional[str] = None,
        osh_diff_scan_after_copr_build: Optional[bool] = True,
        csmock_args: Optional[str] = None,
    ):
        self.config_file_path: Optional[str] = config_file_path
        self.specfile_path: Optional[str] = specfile_path

        self._files_to_sync: list[SyncFilesItem] = files_to_sync or []  # new option
        self._files_to_sync_used: bool = files_to_sync is not None
        self.synced_files: list[SyncFilesItem] = (
            synced_files or []
        )  # old deprecated option
        if synced_files is not None:
            self._warn_user()

        self.patch_generation_ignore_paths = patch_generation_ignore_paths or []
        self.patch_generation_patch_id_digits = patch_generation_patch_id_digits
        self.upstream_project_url: Optional[str] = upstream_project_url
        self.upstream_package_name: Optional[str] = upstream_package_name
        self.paths = paths or ["./"]
        # this is generated by us
        self.downstream_package_name: Optional[str] = downstream_package_name
        self._downstream_project_url: str = downstream_project_url

        # Set up the dist-git instance
        self.dist_git_instance = _construct_dist_git_instance(
            base_url=dist_git_base_url,
            namespace=dist_git_namespace,
            pkg_tool=pkg_tool,
            sig=sig,
        )
        self.pkg_tool = pkg_tool
        self.sig = sig

        self.actions = actions or {}
        self.upstream_ref: Optional[str] = upstream_ref
        self.allowed_gpg_keys = allowed_gpg_keys
        self.create_pr: bool = create_pr
        self.sync_changelog: bool = sync_changelog
        self.create_sync_note: bool = create_sync_note
        self.spec_source_id: str = spec_source_id
        self.notifications = notifications or NotificationsConfig()
        self.identifier = identifier

        # The default is set also on schema level,
        # but for sake of code-generated configs,
        # we want to react on prod events only by default.
        self.packit_instances = (
            packit_instances if packit_instances is not None else [Deployment.prod]
        )

        # template to create an upstream tag name (upstream may use different tagging scheme)
        self.upstream_tag_template = upstream_tag_template
        self.archive_root_dir_template = archive_root_dir_template
        self.copy_upstream_release_description = copy_upstream_release_description
        self.sources = sources or []
        self.merge_pr_in_ci = merge_pr_in_ci
        self.srpm_build_deps = srpm_build_deps
        self.issue_repository = issue_repository
        self.release_suffix = release_suffix
        self.update_release = update_release
        self.upstream_tag_include = upstream_tag_include
        self.upstream_tag_exclude = upstream_tag_exclude
        self.version_update_mask = version_update_mask
        self.prerelease_suffix_pattern = prerelease_suffix_pattern
        self.prerelease_suffix_macro = prerelease_suffix_macro
        self.test_command = test_command or TestCommandConfig()
        self.require = require or RequirementsConfig()
        self.status_name_template = status_name_template
        self.sync_test_job_statuses_with_builds = sync_test_job_statuses_with_builds

        # from deprecated JobMetadataConfig
        self._targets: dict[str, dict[str, Any]]
        if isinstance(_targets, list):
            self._targets = {key: {} for key in _targets}
        else:
            self._targets = _targets or {}
        self.timeout: int = timeout
        self.owner: str = owner
        self.project: str = project
        self.dist_git_branches: Union[
            list[str],
            dict[str, dict[str, list]],
            None,
        ] = (
            dist_git_branches if dist_git_branches else []
        )
        self.branch: str = branch
        self.scratch: bool = scratch
        self.list_on_homepage: bool = list_on_homepage
        self.preserve_project: bool = preserve_project
        self.additional_packages: list[str] = additional_packages or []
        self.additional_repos: list[str] = additional_repos or []
        self.bootstrap = (
            bootstrap if bootstrap is not None else MockBootstrapSetup.default
        )
        self.fmf_url: str = fmf_url
        self.fmf_ref: str = fmf_ref
        self.fmf_path: str = fmf_path
        self.use_internal_tf: bool = use_internal_tf
        self.skip_build: bool = skip_build
        self.env: dict[str, Any] = env or {}
        self.enable_net = enable_net
        self.allowed_pr_authors = (
            allowed_pr_authors if allowed_pr_authors is not None else ["packit"]
        )
        self.allowed_committers = allowed_committers or []
        self.allowed_builders = (
            allowed_builders if allowed_builders is not None else ["packit"]
        )
        self.tmt_plan = tmt_plan
        self.tf_extra_params = tf_extra_params or {}
        self.tf_post_install_script = tf_post_install_script
        self.module_hotfixes = module_hotfixes

        self.image_distribution = image_distribution
        self.image_request = image_request
        self.image_customizations = image_customizations
        self.copr_chroot = copr_chroot

        self.follow_fedora_branching = follow_fedora_branching
        self.upload_sources = upload_sources

        self.parse_time_macros = parse_time_macros or {}

        self.osh_diff_scan_after_copr_build = osh_diff_scan_after_copr_build

        self.csmock_args = csmock_args

    @property
    def dist_git_base_url(self) -> str:
        return self.dist_git_instance.url

    @property
    def dist_git_namespace(self) -> Optional[str]:
        return self.dist_git_instance.namespace

    @property
    def targets_dict(self):
        return self._targets

    @property
    def targets(self):
        """For backward compatibility."""
        return set(self._targets.keys())

    def _warn_user(self):
        logger = logging.getLogger(__name__)
        msg = "synced_files option is deprecated. Use files_to_sync option instead."
        logger.warning(msg)
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        if self._files_to_sync_used:
            logger.warning(
                "You are setting both files_to_sync and synced_files."
                " Packit will use files_to_sync. You should remove "
                "synced_files since it is deprecated.",
            )

    @property
    def files_to_sync(self) -> list[SyncFilesItem]:
        """
        synced_files is the old option we want to deprecate.
        Spec file and configuration file can be automatically added to
        the list of synced_files (see get_all_files_to_sync method)

        files_to_sync is the new option. Files to be synced are just those listed here.
        Spec file and configuration file will not be automatically added
        to the list of files_to_sync (see get_all_files_to_sync method).

        files_to_sync has precedence over synced_files.

        Once the old option will be removed this method can be removed as well.
        """

        if self._files_to_sync_used:
            return self._files_to_sync
        if self.synced_files:
            return self.synced_files
        return []

    def __repr__(self):
        # required to avoid cyclical imports
        from packit.schema import CommonConfigSchema

        s = CommonConfigSchema()
        # For __repr__() return a JSON-encoded string, by using dumps().
        # Mind the 's'!
        return f"CommonPackageConfig: {s.dumps(self)}"

    def __eq__(self, other: object):
        if not isinstance(other, CommonPackageConfig):
            raise PackitConfigException(
                "Provided object is not a CommonPackageConfig instance.",
            )
        # required to avoid cyclical imports
        from packit.schema import CommonConfigSchema

        s = CommonConfigSchema()
        # Compare the serialized objects.
        return s.dump(self) == s.dump(other)

    @property
    def downstream_project_url(self) -> str:
        if not self._downstream_project_url:
            self._downstream_project_url = self.dist_git_package_url
        return self._downstream_project_url

    @property
    def dist_git_package_url(self):
        if self.downstream_package_name:
            return (
                f"{self.dist_git_base_url}{self.dist_git_namespace}/"
                f"{self.downstream_package_name}.git"
            )
        return None

    @property
    def spec_source_id_number(self) -> int:
        """
        Return spec_source_id as a number, reverse of spec_source_id_fm
        """
        return int(next(iter(split(r"(\d+)", self.spec_source_id)[1:]), 0))

    def get_specfile_sync_files_item(self, from_downstream: bool = False):
        """
        Get SyncFilesItem object for the specfile.
        :param from_downstream: True when syncing from downstream
        :return: SyncFilesItem
        """
        upstream_specfile_path = self.specfile_path
        downstream_specfile_path = (
            f"{self.downstream_package_name}.spec"
            if self.downstream_package_name
            else basename(upstream_specfile_path)
        )
        return SyncFilesItem(
            src=[
                downstream_specfile_path if from_downstream else upstream_specfile_path,
            ],
            dest=upstream_specfile_path
            if from_downstream
            else downstream_specfile_path,
        )

    def get_all_files_to_sync(self):
        """
        Adds the default files (config file, spec file) to synced files
        if the new files_to_sync option is not yet used otherwise
        do not performer any addition to the list of files to be synced.

        When the old option will be removed this method could be removed as well.

        :return: Files to be synced
        """
        files = self.files_to_sync

        if not self._files_to_sync_used:
            if self.specfile_path not in iter_srcs(files):
                files.append(self.get_specfile_sync_files_item())

            if self.config_file_path and self.config_file_path not in iter_srcs(files):
                # this relative because of glob: "Non-relative patterns are unsupported"
                files.append(
                    SyncFilesItem(
                        src=[self.config_file_path],
                        dest=self.config_file_path,
                    ),
                )

        return files


class MultiplePackages:
    """
    Base class for configuration classes which have a "packages" attribute

    TODO: remove this once support to access the attributes of a single package
          is dropped.

    Notes:
        1. Need to define '__setattr__', b/c some properties of the config
           objects are set here and there. This makes it necessary to use the
           'super().__setattr__()' syntax when writing attributes of this
           class.
        2. In theory accessing attributes of this class with 'self.<attribute>'
           works, until it doesn't. Objects being deepcopied is one such
           case. Using 'self.__getattribute__("attribute")' solves this issue.
    """

    up_url_key = "upstream_project_url"

    def __init__(self, packages: dict[str, CommonPackageConfig]):
        super().__setattr__("packages", packages)
        super().__setattr__("_first_package", next(iter(packages)))

    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        if len(self.__getattribute__("packages")) == 1:
            package = self.__getattribute__("packages")[
                self.__getattribute__("_first_package")
            ]
            return getattr(package, name)
        if name == self.up_url_key:
            package = self.__getattribute__("packages")[
                self.__getattribute__("_first_package")
            ]
            return getattr(package, name)
        raise AttributeError(
            f"It is ambiguous to get {name}: "
            "there is more than one package in the config.",
        )

    def __setattr__(self, name, value):
        if name in self.__dict__ or "packages" not in self.__dict__:
            super().__setattr__(name, value)
        elif len(self.__getattribute__("packages")) == 1:
            package = self.__getattribute__("packages")[
                self.__getattribute__("_first_package")
            ]
            setattr(package, name, value)
        elif name == self.up_url_key:
            for package in self.__getattribute__("packages").values():
                setattr(package, name, value)
        else:
            raise AttributeError(
                f"It is ambiguous to set {name}: "
                "there is more than one package in the config.",
            )

    def get_package_names_as_env(self) -> dict[str, str]:
        """Creates a dict with package_name,
        downstream_package_name and upstream_package_name.

        If the config contains multiple packages
        raise an Exception.
        """
        if len(self.packages) == 1:
            for packit_package_name, package_config in self.packages.items():
                env = {}
                env["PACKIT_CONFIG_PACKAGE_NAME"] = packit_package_name
                env["PACKIT_UPSTREAM_PACKAGE_NAME"] = (
                    package_config.upstream_package_name or ""
                )
                env["PACKIT_DOWNSTREAM_PACKAGE_NAME"] = (
                    package_config.downstream_package_name or ""
                )
                return env
            raise PackitConfigException("No packages in config")
        raise PackitConfigException("Multiple packages in config")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import sys

from flexmock import flexmock

from packit.cli.utils import cover_packit_exception
from packit.exceptions import PackitException


def test_cover_packit_exception_decorator():
    flexmock(sys).should_receive("exit")

    @cover_packit_exception
    def covered_func():
        raise PackitException("Test exception")

    covered_func()


def test_cover_packit_exception_decorator_attribute():
    flexmock(sys).should_receive("exit")

    @cover_packit_exception
    def covered_func(attribute):
        assert attribute
        raise PackitException("Other test exception")

    covered_func(attribute=flexmock())


def test_cover_packit_exception_decorator_exit_code_default():
    flexmock(sys).should_receive("exit").with_args(2)

    @cover_packit_exception()
    def covered_func():
        raise PackitException("Test exception")

    covered_func()


def test_cover_packit_exception_decorator_exit_code_keyboard_interrupt():
    flexmock(sys).should_receive("exit").with_args(1)

    @cover_packit_exception()
    def covered_func():
        raise KeyboardInterrupt("Test keyboard interrupt")

    covered_func()


def test_cover_packit_exception_decorator_exit_code_other_exception():
    flexmock(sys).should_receive("exit").with_args(4)

    class CustomException(Exception):
        pass

    @cover_packit_exception()
    def covered_func():
        raise CustomException("Test keyboard interrupt")

    covered_func()


def test_cover_packit_exception_decorator_exit_code_other_exception_override():
    flexmock(sys).should_receive("exit").with_args(5)

    class CustomException(Exception):
        pass

    @cover_packit_exception(exit_code=5)
    def covered_func():
        raise CustomException("Test keyboard interrupt")

    covered_func()


def test_cover_packit_exception_decorator_exit_code_override():
    flexmock(sys).should_receive("exit").with_args(5)

    @cover_packit_exception(exit_code=5)
    def covered_func():
        raise PackitException("Test exception")

    covered_func()


def test_cover_packit_exception_decorator_config_debug_false():
    flexmock(sys).should_receive("exit")

    @cover_packit_exception
    def covered_func(config=None):
        raise PackitException("Test exception")

    covered_func(config=flexmock(debug=False))


def test_cover_packit_exception_decorator_config_debug_true():
    flexmock(sys).should_receive("exit")

    @cover_packit_exception
    def covered_func(config=None):
        raise PackitException("Test exception")

    covered_func(config=flexmock(debug=True))


def test_cover_packit_exception_decorator_other_exception():
    flexmock(sys).should_receive("exit").with_args(4)

    class CustomException(Exception):
        pass

    @cover_packit_exception
    def covered_func():
        raise CustomException("Other test exception")

    covered_func()


def test_cover_packit_exception_decorator_other_exception_config_debug():
    flexmock(sys).should_receive("exit")

    class CustomException(Exception):
        pass

    @cover_packit_exception
    def covered_func(config=None):
        raise CustomException("Other test exception")

    covered_func(config=flexmock(debug=True))
